/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 * Created on March 25 2003
 */
package org.jboss.cache;


import org.jboss.cache.lock.IdentityLock;
import org.jboss.cache.lock.LockingException;
import org.jboss.cache.lock.TimeoutException;
import org.jboss.logging.Logger;

import java.io.Serializable;
import java.util.*;

/**
 * Represents a node in the tree. Has a relative name and a Fqn. Maintains a hashmap.
 * If the node is created in a replicated cache, the relative and fully qualified name,
 * and the keys and values of the hashmap have to be serializable.
 * <p>Note that current version supports different levels of transaction locking such
 * as simple locking (<CODE>TreeCache.SERIALIZABLE</code>, or R/W lock with upgrade (
 * <code>REPEATABLE_READ</code>) --that is the read lock
 * will be automatically upgraded to write lock when the same owner intends to modify
 * the data after read.</p>
 * <br/>
 * todo: implement Externalizable
 *
 * @author Bela Ban March 25 2003
 * @version $Revision: 1.24.2.2 $
 */
public class Node implements Serializable
{
   /**
    * Relative name of the node, e.g. "c"
    */
   Object name = null;

   /**
    * Fully qualified name (e.g. "/a/b/c")
    */
   Fqn fqn = null;

   /**
    * Parent node. For navigational purposes
    */
   Node parent = null;

   /**
    * Children. Keys: child name (Object), value: Node
    */
   HashMap children = null;

   /**
    * Data for the current node. Keys: Object, value: Object
    */
   HashMap data = null;

   /**
    * Various annotations, e.g. the lock for this node, whether the node
    * is transient or replicated, etc
    */
   transient Map props = null;

   /**
    * Lock manager that manages locks to be acquired when accessing
    * the node inside a transaction. Lazy loading just in case locking is not needed
    */
   transient IdentityLock lock_ = null;

   /**
    * Declare locking type with the intention to read or write data.
    */
   public static final int LOCK_TYPE_READ = 1;
   public static final int LOCK_TYPE_WRITE = 2;

   static transient Logger log = Logger.getLogger(Node.class);

   protected final int INDENT = 4;

   public Node(Object child_name, Fqn fqn, Node parent, Map data)
   {
      name = child_name;
      this.fqn = fqn;
      this.parent = parent;
      if (data != null)
      {
         this.data = new HashMap();
         this.data.putAll(data);
      }
   }

   public Node(Object child_name, Fqn fqn, Node parent, Object key, Object value)
   {
      name = child_name;
      this.fqn = fqn;
      this.parent = parent;
      if (data == null) data = new HashMap();
      data.put(key, value);
   }

   public Object getName()
   {
      return name;
   }

   public Fqn getFqn()
   {
      return fqn;
   }

   public Object get(Object key)
   {
      return data != null ? data.get(key) : null;
   }

   public Node getChild(Object child_name)
   {
      return child_name == null ? null : children == null ? null : (Node) children.get(child_name);
   }

   public Node getParent()
   {
      return parent;
   }

   public Set getDataKeys()
   {
      return data != null ? data.keySet() : null;
   }

   public Map getProperties()
   {
      return props;
   }

   public boolean childExists(Object child_name)
   {
      if (child_name == null) return false;
      return children != null ? children.containsKey(child_name) : false;
   }

   public void put(Map data, boolean erase)
   {
      if (erase && this.data != null)
         this.data.clear();
      if (data == null) return;
      if (this.data == null)
         this.data = new HashMap();
      this.data.putAll(data);
   }

   public void put(Map data)
   {
      put(data, false);
   }

   void putInternal(HashMap data)
   {
      this.data = data;
   }

   public Object put(Object key, Object value)
   {
      if (this.data == null)
         this.data = new HashMap();
      return this.data.put(key, value);
   }

   public synchronized Node createChild(Object child_name, Fqn fqn, Node parent)
   {
      Node child = null;

      if (child_name == null) return null;
      if (children == null) children = new HashMap();
      child = (Node) children.get(child_name);
      if (child == null)
      {
         child = new Node(child_name, fqn, parent, null);
         children.put(child_name, child);
      }
      if (log.isDebugEnabled())
         log.debug("createChild: fqn=" + fqn);
      return child;
   }

   public synchronized Node createChild(Object child_name, Fqn fqn, Node parent, Map data)
   {
      Node child = null;

      if (child_name == null) return null;
      if (children == null) children = new HashMap();
      child = (Node) children.get(child_name);
      if (child != null)
         child.put(data);
      else
      {
         child = new Node(child_name, fqn, parent, data);
         children.put(child_name, child);
      }
      if (log.isDebugEnabled())
         log.debug("createChild: fqn=" + fqn);
      return child;
   }

   public synchronized Node createChild(Object child_name, Fqn fqn, Node parent, Object key, Object value)
   {
      Node child = null;

      if (child_name == null) return null;
      if (children == null) children = new HashMap();
      child = (Node) children.get(child_name);
      if (child != null)
         child.put(key, value);
      else
      {
         child = new Node(child_name, fqn, parent, key, value);
         children.put(child_name, child);
      }
      if (log.isDebugEnabled())
         log.debug("createChild: fqn=" + fqn);
      return child;
   }

   /**
    * Adds the (already created) child node. Replaces existing node if present.
    *
    * @param child_name
    * @param n
    */
   void addChild(Object child_name, Node n)
   {
      if (child_name == null)
         return;
      children.put(child_name, n);
   }

   public Object remove(Object key)
   {
      return data != null ? data.remove(key) : null;
   }

   public void clear()
   {
      if (data != null)
         data.clear();
      data = null;
   }

   public void removeChild(Object child_name)
   {
      if (children != null)
         children.remove(child_name);
   }

   public void removeAllChildren()
   {
      if (children != null)
      {
         children.clear();
         children = null;
      }
   }

   public void print(StringBuffer sb, int indent)
   {
      printIndent(sb, indent);
      sb.append(TreeCache.SEPARATOR).append(name);
      if (children != null && children.size() > 0)
      {
         Collection values = children.values();
         for (Iterator it = values.iterator(); it.hasNext();)
         {
            sb.append("\n");
            ((Node) it.next()).print(sb, indent + INDENT);
         }
      }
   }

   public void printDetails(StringBuffer sb, int indent)
   {
      Map.Entry entry;
      printIndent(sb, indent);
      sb.append(TreeCache.SEPARATOR).append(name);
      if (data != null)
      {
         sb.append("\n");
         for (Iterator it = data.entrySet().iterator(); it.hasNext();)
         {
            entry = (Map.Entry) it.next();
            sb.append(entry.getKey()).append(": ").append(entry.getValue()).append("\n");
         }
      }
      if (children != null && children.size() > 0)
      {
         Collection values = children.values();
         for (Iterator it = values.iterator(); it.hasNext();)
         {
            sb.append("\n");
            ((Node) it.next()).printDetails(sb, indent);
         }
      }
   }

   public void printLockInfo(StringBuffer sb, int indent)
   {
      boolean locked = lock_ != null ? lock_.isReadLocked() : false;
      Collection owners = lock_ != null ? lock_.getReaderOwners() : null;
      printIndent(sb, indent);
      sb.append(TreeCache.SEPARATOR).append(name);
      sb.append(" \t(locked=").append(locked).append(", read owner=").append(owners).append(")");
      locked = lock_ != null ? lock_.isReadLocked() : false;
      Object owner = lock_ != null ? lock_.getWriterOwner() : null;
      printIndent(sb, indent);
      sb.append(TreeCache.SEPARATOR).append(name);
      sb.append(" \t(locked=").append(locked).append(", write owner=").append(owner).append(")");
      if (children != null && children.size() > 0)
      {
         Collection values = children.values();
         for (Iterator it = values.iterator(); it.hasNext();)
         {
            sb.append("\n");
            ((Node) it.next()).printLockInfo(sb, indent + INDENT);
         }
      }
   }

   public void printIndent(StringBuffer sb, int indent)
   {
      if (sb != null)
      {
         for (int i = 0; i < indent; i++)
            sb.append(" ");
      }
   }

   public String toString()
   {
      StringBuffer sb = new StringBuffer();
      if (name != null) sb.append("\nname=" + name);
      if (fqn != null) sb.append("\nfqn=" + fqn);
      if (data != null) sb.append("\ndata=" + data);
      if (lock_ != null)
      {
         sb.append("\n read locked=").append(isReadLocked());
         sb.append("\n write locked=").append(isWriteLocked());
//            sb.append(" (num permits=").append(getNumPermits()).append(")");
      }
      return sb.toString();
   }


   public Object clone() throws CloneNotSupportedException
   {
      Node n = new Node(name, fqn, parent != null ? (Node) parent.clone() : null, data);
      if (children != null) n.children = (HashMap) children.clone();
      return n;
   }

   /**
    * Acquire a node object by locking it first. The lock type is specified via
    * <code>lock_type</code>.
    *
    * @param owner     Tansaction context owner.
    * @param timeout   Timeout in milliseconds
    * @param lock_type @see Node
    * @throws LockingException
    * @throws TimeoutException
    */
   public void acquire(Object owner, long timeout, int lock_type)
      throws LockingException, TimeoutException
   {
      if (lock_type == LOCK_TYPE_READ)
      {
         acquireReadLock(owner, timeout);
      }
      else
      {
         acquireWriteLock(owner, timeout);
      }
   }

   protected void acquireReadLock(Object owner, long timeout)
      throws LockingException, TimeoutException
   {
      if (lock_ == null)
      {
         synchronized (this)
         {
            if (lock_ == null)
               lock_ = new IdentityLock();
         }
      }
      lock_.acquireReadLock(owner, timeout);
      if (log.isDebugEnabled())
         log.debug("acquired RL: fqn=" + fqn + ", owner=" + owner + ", lock=" + lock_);
   }

   protected void acquireWriteLock(Object owner, long timeout)
      throws LockingException, TimeoutException
   {
      if (lock_ == null)
      {
         synchronized (this)
         {
            if (lock_ == null)
               lock_ = new IdentityLock();
         }
      }
      lock_.acquireWriteLock(owner, timeout);
      if (log.isDebugEnabled())
         log.debug("acquired WL: fqn=" + fqn + ", owner=" + owner + ", lock=" + lock_);
   }

   /**
    * Acquires locks for the entire subtree
    */
   protected void acquireAll(Object owner, long timeout, int lock_type)
      throws LockingException, TimeoutException
   {
      Node tmp;
      acquire(owner, timeout, lock_type);
      if (children != null)
      {
         for (Iterator it = children.values().iterator(); it.hasNext();)
         {
            tmp = (Node) it.next();
            tmp.acquireAll(owner, timeout, lock_type);
         }
      }
   }

   protected void release(Object owner)
   {
      if (log.isDebugEnabled())
         log.debug("release(): fqn=" + fqn + ", owner=" + owner);
      if (lock_ == null)
      {
         synchronized (this)
         {
            if (lock_ == null)
               return;
            else
               lock_.release(owner);
         }
      }
      else
      {
         lock_.release(owner);
      }
   }

   protected void releaseForce()
   {
      synchronized (this)
      {
         if (lock_ != null)
            lock_.releaseForce();
      }
   }

   /**
    * Releases locks for the entire subtree
    */
   protected void releaseAll(Object owner)
   {
      Node tmp;
      if (children != null)
      {
         for (Iterator it = children.values().iterator(); it.hasNext();)
         {
            tmp = (Node) it.next();
            tmp.releaseAll(owner);
         }
      }
      release(owner);
   }

   /**
    * Releases locks for the entire subtree. Forces release regardless of owner
    */
   protected void releaseAllForce()
   {
      Node tmp;
      if (children != null)
      {
         for (Iterator it = children.values().iterator(); it.hasNext();)
         {
            tmp = (Node) it.next();
            tmp.releaseAllForce();
         }
      }
      releaseForce();
   }

   protected boolean isReadLocked()
   {
      return lock_ != null ? lock_.isReadLocked() : false;
   }

   protected boolean isWriteLocked()
   {
      return lock_ != null ? lock_.isWriteLocked() : false;
   }

/*
    protected long getNumPermits() {
        return lock_ != null? lock_.getNumPermits() : 0;
    }
    */

   public IdentityLock getLock()
   {
      return lock_;
   }

   public Map getChildren()
   {
      return children;
   }

   public Map getData()
   {
      if (data == null) return null;

      Map tmp = new HashMap(data);
      return tmp;
   }

   public boolean hasChild()
   {
      if (children == null || children.size() == 0) return false;

      return true;
   }
}

