/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache;


import org.jboss.logging.Logger;

import java.lang.reflect.Method;
import java.util.*;


/**
 * Periodically (or when certain size is exceeded) takes elements and replicates them.
 *
 * @author <a href="mailto:bela@jboss.org">Bela Ban</a> May 24, 2003
 * @version $Revision: 1.3.4.2 $
 */
public class ReplicationQueue
{
   /**
    * We flush every 5 seconds. Inactive if -1 or 0
    */
   long interval = 5000;

   /**
    * Max elements before we flush
    */
   long max_elements = 500;

   /**
    * Holds the replication jobs
    */
   LinkedList elements = new LinkedList();

   /**
    * For periodical replication
    */
   Timer timer = null;

   /**
    * The timer task, only calls flush() when executed by Timer
    */
   MyTask task = null;

   Logger log = Logger.getLogger(ReplicationQueue.class);


   public ReplicationQueue()
   {

   }

   public ReplicationQueue(long interval, long max_elements)
   {
      this.interval = interval;
      this.max_elements = max_elements;
   }


   public long getInterval()
   {
      return interval;
   }

   public void setInterval(long interval)
   {
      this.interval = interval;
      stop();
      start();
   }

   public long getMax_elements()
   {
      return max_elements;
   }

   public void setMax_elements(long max_elements)
   {
      this.max_elements = max_elements;
   }

   public void start()
   {
      if (interval > 0)
      {
         if (task == null)
            task = new MyTask();
         if (timer == null)
         {
            timer = new Timer(true);
            timer.schedule(task,
               500, // delay before initial flush
               interval); // interval between flushes
         }
      }
   }


   public void stop()
   {
      if (task != null)
      {
         task.cancel();
         task = null;
      }
      if (timer != null)
      {
         timer.cancel();
         timer = null;
      }
   }


   public void add(Job job)
   {
      synchronized (elements)
      {
         if (job != null)
            elements.add(job);
         if (elements.size() >= max_elements)
            flush();
      }
   }


   public void flush()
   {
      synchronized (elements)
      {
         if (log.isInfoEnabled())
            log.info("flush(): flushing repl queue (num elements=" + elements.size() + ")");
         for (Iterator it = elements.iterator(); it.hasNext();)
         {
            Job j = (Job) it.next();
            try
            {
               j.cache.callRemoteMethods(j.members, j.m, j.args, false, j.exclude_self, 5000);
            }
            catch (Exception e)
            {
               log.error("flush(): failed replicating job " + j, e);
            }
         }
         elements.clear();
      }
   }


   public class MyTask extends TimerTask
   {
      public void run()
      {
         flush();
      }
   }


   public static class Job
   {
      TreeCache cache;
      Vector members = null;
      boolean exclude_self = true;
      Method m;
      Object[] args;

      public Job(TreeCache cache, Vector members, boolean exclude_self, Method m, Object[] args)
      {
         this.cache = cache;
         this.members = members; // must already be cloned
         this.exclude_self = exclude_self;
         this.m = m;
         this.args = args;
      }

      public String toString()
      {
         StringBuffer sb = new StringBuffer();
         sb.append("method=").append(m != null ? m.getName() : "null");
         sb.append(", args=").append(args);
         return sb.toString();
      }
   }

}
