/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache;

import org.jboss.cache.lock.*;
import org.jboss.cache.lock.TimeoutException;
import org.jboss.cache.eviction.LRUPolicy;
import org.jboss.system.ServiceMBeanSupport;
import org.jboss.util.NestedRuntimeException;
import org.jgroups.*;
import org.jgroups.blocks.GroupRequest;
import org.jgroups.blocks.MethodCall;
import org.jgroups.blocks.RpcDispatcher;
import org.jgroups.util.Rsp;
import org.jgroups.util.RspList;
import org.jgroups.util.Util;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;

import javax.transaction.Status;
import javax.transaction.Synchronization;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.*;


/**
 * A tree-like structure that is replicated across several members. Updates will
 * be multicast to all group members reliably and in the same order. User has the
 * option to set transaction locking level now (e.g., <code>SERIALIZABLE</code>, or
 * <code>REPEATABLE_READ</code>.
 *
 * @author Bela Ban March 25 2003
 * @author Ben Wang July 2003
 * @version $Id: TreeCache.java,v 1.79.2.8 2004/05/19 03:08:09 belaban Exp $
 * @jmx.mbean extends="org.jboss.system.ServiceMBean"
 * <p/>
 * todo: emit notifications only on commit. bad case: emit notif about node creation,
 * // then rollback TX
 */
public class TreeCache extends ServiceMBeanSupport implements TreeCacheMBean, Cloneable, MembershipListener
{
   protected Node root = new Node(SEPARATOR, Fqn.fromString(SEPARATOR), null, null);
   protected Vector listeners = new Vector();
   protected JChannel channel = null;
   protected String cluster_name = "TreeCache-Group";
   protected String cluster_props = null;
   protected Vector members = new Vector();
   protected RpcDispatcher disp = null;
   protected LockManager lock_mgr = new LockManagerImpl();
   protected MessageListener ml = new MessageListenerAdaptor();
   protected long state_fetch_timeout = 20000;
   protected long sync_repl_timeout = 30000;
   protected boolean use_repl_queue = false;
   protected int repl_queue_max_elements = 1000;
   protected long repl_queue_interval = 5000;

   /**
    * only used internally, no setters or getters
    */
   protected boolean locking = true;

   protected boolean fetch_state_on_startup = true;
   protected long lock_acquisition_timeout = 10000;
   protected long lock_lease_timeout = 60000;
   protected int max_capacity = 100000;
   protected String eviction_policy_class = null;
   protected TreeCacheListener eviction_policy_provider_ = null;
   protected int cache_mode = LOCAL;
   protected Method put_data_method = null;
   protected Method put_data_erase_method = null;
   protected Method put_key_val_method = null;
   protected Method remove_node_method = null;
   protected Method remove_key_method = null;
   protected Method remove_data_method = null;
   protected Method prepare_method = null;
   protected Method commit_method = null;
   protected Method rollback_method = null;
   protected Method addChild_method = null;
   protected boolean isStateSet = false;
   protected Object stateLock = new Object();
   // Distinguish the node operation type. Use for locking granularity
   public static final int NODE_READ = 1;
   public static final int NODE_WRITE = 2;
   protected IsolationLevel isolationLevel = IsolationLevel.REPEATABLE_READ;
   protected Map thread2GtxMap_ = new HashMap();
   // Eviction policy configuration in xml Element
   protected Element evictConfig_ = null;

   /**
    * Maintains (a) mapping between Transaction and GlobalTransactions and
    * (b) between GlobalTransactions and TransactionEntries
    */
   protected TransactionTable tx_table = new TransactionTable();


   /**
    * Method to acquire a TransactionManager. By default we use JBossTransactionManagerLookup. Has
    * to be set before calling {@link #start()}
    */
   protected TransactionManagerLookup tm_lookup = null;

   /**
    * Class of the implementation of TransactionManagerLookup
    */
   protected String tm_lookup_class = null;

   /**
    * Used to get the Transaction associated with the current thread
    */
   protected TransactionManager tm = null;

   /**
    * Queue used to replicate updates when mode is repl-async
    */
   protected ReplicationQueue repl_queue = null;

   public static final String SEPARATOR = "/";

   /**
    * Entries in the cache are by default local; ie. not replicated
    */
   public static final int LOCAL = 1;

   /**
    * Entries in the cache are by default replicated (asynchronously)
    */
   public static final int REPL_ASYNC = 2;

   /**
    * Entries in the cache are by default replicated (synchronously)
    */
   public static final int REPL_SYNC = 3;


   /**
    * Creates a channel with the given properties. Connects to the channel, then creates a PullPushAdapter
    * and starts it
    */
   public TreeCache(String cluster_name, String props, long state_fetch_timeout) throws Exception
   {
      super();
      initMethods();
      if (cluster_name != null)
         this.cluster_name = cluster_name;
      if (props != null)
         this.cluster_props = props;
      this.state_fetch_timeout = state_fetch_timeout;
   }

   public TreeCache() throws Exception
   {
      super();
      initMethods();
   }

   /**
    * Expects an already connected channel. Creates a PullPushAdapter and starts it
    */
   public TreeCache(JChannel channel) throws Exception
   {
      super();
      initMethods();
      this.channel = channel;
   }

   /**
    * @return
    * @jmx.managed-attribute access="read-only"
    */
   public Object getLocalAddress()
   {
      return channel != null ? channel.getLocalAddress() : null;
   }

   /**
    * @return
    * @jmx.managed-attribute access="read-only"
    */
   public Vector getMembers()
   {
      return members;
   }

   /**
    * Get the name of the replication group
    *
    * @jmx.managed-attribute
    */
   public String getClusterName()
   {
      return cluster_name;
   }

   /**
    * Set the name of the replication group
    *
    * @jmx.managed-attribute
    */
   public void setClusterName(String name)
   {
      cluster_name = name;
   }

   /**
    * Get the cluster properties (e.g. the protocol stack specification in case of JGroups)
    *
    * @jmx.managed-attribute
    */
   public String getClusterProperties()
   {
      return cluster_props;
   }

   /**
    * Set the cluster properties. If the cache is to use the new properties,
    * it has to be redeployed
    *
    * @param cluster_props The properties for the cluster (JGroups)
    * @jmx.managed-attribute
    */
   public void setClusterProperties(String cluster_props)
   {
      this.cluster_props = cluster_props;
   }

   /**
    * Setup eviction policy configuration
    *
    * @jmx.managed-attribute access="write-only"
    */
   public void setEvictionPolicyConfig(Element config)
   {
      evictConfig_ = config;
      log.info("setEvictionPolicyConfig(): " + config);
   }

   /**
    */
   public Element getEvictionPolicyConfig()
   {
      return evictConfig_;
   }

   /**
    * Convert a list of elements to the JG property string
    *
    * @jmx.managed-attribute access="write-only"
    */
   public void setClusterConfig(Element config)
   {
      StringBuffer buffer = new StringBuffer();
      NodeList stack = config.getChildNodes();
      int length = stack.getLength();

      for (int s = 0; s < length; s++)
      {
         org.w3c.dom.Node node = stack.item(s);
         if (node.getNodeType() != org.w3c.dom.Node.ELEMENT_NODE)
            continue;

         Element tag = (Element) node;
         String protocol = tag.getTagName();
         buffer.append(protocol);
         NamedNodeMap attrs = tag.getAttributes();
         int attrLength = attrs.getLength();
         if (attrLength > 0)
            buffer.append('(');
         for (int a = 0; a < attrLength; a++)
         {
            Attr attr = (Attr) attrs.item(a);
            String name = attr.getName();
            String value = attr.getValue();
            buffer.append(name);
            buffer.append('=');
            buffer.append(value);
            if (a < attrLength - 1)
               buffer.append(';');
         }
         if (attrLength > 0)
            buffer.append(')');
         buffer.append(':');
      }
      // Remove the trailing ':'
      buffer.setLength(buffer.length() - 1);
      setClusterProperties(buffer.toString());
      log.info("setClusterConfig(): setting cluster properties " +
         "from xml to: " + cluster_props);
   }


   /**
    * Get the maximum number of entries (cache eviction kicks in when reached)
    *
    * @jmx.managed-attribute
    */
   public int getMaxCapacity()
   {
      return max_capacity;
   }

   /**
    * Set the maximum number of entries (cache eviction kicks in when reached)
    *
    * @jmx.managed-attribute
    */
   public void setMaxCapacity(int max_capacity)
   {
      this.max_capacity = max_capacity;
   }

   /**
    * Get the max time to wait until the initial state is retrieved.
    * This is used in a replicating cache: when a new cache joins the cluster,
    * it needs to acquire the (replicated) state of the other members to
    * initialize itself. If no state has been received within <tt>timeout</tt>
    * milliseconds, the map will be empty.
    *
    * @return long Number of milliseconds to wait for the state. 0 means to
    *         wait forever.
    * @jmx.managed-attribute
    */
   public long getInitialStateRetrievalTimeout()
   {
      return state_fetch_timeout;
   }

   /**
    * Set the initial state transfer timeout
    * (see {@link #getInitialStateRetrievalTimeout()})
    *
    * @jmx.managed-attribute
    */
   public void setInitialStateRetrievalTimeout(long timeout)
   {
      state_fetch_timeout = timeout;
   }

   /**
    * Returns the current caching mode. Valid values are
    * <ul>
    * <li>LOCAL
    * <li>REPL_ASYNC
    * <li>REPL_SYNC
    * <ul>
    *
    * @return String The caching mode
    * @jmx.managed-attribute
    */
   public String getCacheMode()
   {
      return mode2String(cache_mode);
   }

   private String mode2String(int mode)
   {
      switch (mode)
      {
         case LOCAL:
            return "LOCAL";
         case REPL_ASYNC:
            return "REPL_ASYNC";
         case REPL_SYNC:
            return "REPL_SYNC";
         default:
            throw new RuntimeException("setCacheMode(): caching mode " + mode + " is invalid");
      }
   }

   /**
    * Sets the default caching mode)
    *
    * @jmx.managed-attribute
    */
   public void setCacheMode(String mode) throws Exception
   {
      int m = string2Mode(mode);
      setCacheMode(m);
   }


   /**
    * Sets the default cache mode. Valid arguments are
    * <ol>
    * <li>TreeCache.LOCAL
    * <li>TreeCache.REPL_ASYNC
    * <li>TreeCache.REPL_SYNC
    * </ol>
    *
    * @param mode
    */
   public void setCacheMode(int mode) throws Exception
   {
      if (mode == LOCAL || mode == REPL_ASYNC || mode == REPL_SYNC)
         this.cache_mode = mode;
      else
         throw new Exception("setCacheMode(): caching mode " + mode + " is invalid");
   }


   /**
    * Returns the default max timeout after which synchronous replication calls
    * return.
    *
    * @return long Number of milliseconds after which a sync repl call must
    *         return. 0 means to wait forever
    * @jmx.managed-attribute
    */
   public long getSyncReplTimeout()
   {
      return sync_repl_timeout;
   }

   /**
    * Sets the default maximum wait time for synchronous replication to receive all
    * results
    *
    * @jmx.managed-attribute
    */
   public void setSyncReplTimeout(long timeout)
   {
      sync_repl_timeout = timeout;
   }

   /**
    * @return
    * @jmx.managed-attribute
    */
   public boolean getUseReplQueue()
   {
      return use_repl_queue;
   }

   /**
    * @param flag
    * @jmx.managed-attribute
    */
   public void setUseReplQueue(boolean flag)
   {
      use_repl_queue = flag;
      if (flag)
      {
         if (repl_queue == null)
         {
            repl_queue = new ReplicationQueue(repl_queue_interval, repl_queue_max_elements);
            if (repl_queue_interval >= 0)
               repl_queue.start();
         }
      }
      else
      {
         if (repl_queue != null)
         {
            repl_queue.stop();
            repl_queue = null;
         }
      }
   }

   /**
    * @return
    * @jmx.managed-attribute
    */
   public long getReplQueueInterval()
   {
      return repl_queue_interval;
   }

   /**
    * @param interval
    * @jmx.managed-attribute
    */
   public void setReplQueueInterval(long interval)
   {
      this.repl_queue_interval = interval;
      if (repl_queue != null)
      {
         repl_queue.setInterval(interval);
      }
   }

   /**
    * @return
    * @jmx.managed-attribute
    */
   public int getReplQueueMaxElements()
   {
      return repl_queue_max_elements;
   }

   /**
    * @param max_elements
    * @jmx.managed-attribute
    */
   public void setReplQueueMaxElements(int max_elements)
   {
      this.repl_queue_max_elements = max_elements;
      if (repl_queue != null)
         repl_queue.setMax_elements(max_elements);
   }


   /**
    * Returns the transaction isolation level.
    *
    * @jmx.managed-attribute
    */
   public String getIsolationLevel()
   {
      return isolationLevel.toString();
   }

   /**
    * Set the transaction isolation level. This determines the locking strategy to be used
    *
    * @jmx.managed-attribute
    */
   public void setIsolationLevel(String level) throws Exception
   {
      IsolationLevel tmp_level = IsolationLevel.stringToIsolationLevel(level);

      if (tmp_level == null)
      {
         throw new Exception("TreeCache.setIsolationLevel(): level \"" + level + "\" is invalid");
      }
      setIsolationLevel(tmp_level);
   }

   /**
    * @param level
    */
   public void setIsolationLevel(IsolationLevel level)
   {
      isolationLevel = level;
      if (isolationLevel != IsolationLevel.NONE)
         locking = true;
      else
         locking = false;
      LockStrategyFactory.setIsolationLevel(level);
   }

   /**
    * @return
    * @jmx.managed-attribute
    */
   public boolean getFetchStateOnStartup()
   {
      return fetch_state_on_startup;
   }


   /**
    * @param flag
    * @jmx.managed-attribute
    */
   public void setFetchStateOnStartup(boolean flag)
   {
      fetch_state_on_startup = flag;
   }


   /**
    * Default max time to wait for a lock. If the lock cannot be acquired
    * within this time, a LockingException will be thrown.
    *
    * @return long Max number of milliseconds to wait for a lock to be acquired
    * @jmx.managed-attribute
    */
   public long getLockAcquisitionTimeout()
   {
      return lock_acquisition_timeout;
   }

   /**
    * Set the max time for lock acquisition. A value of 0 means to wait forever (not recomended).
    * Note that lock acquisition timeouts may be removed in the future when we
    * have deadlock detection.
    *
    * @param timeout
    * @jmx.managed-attribute
    */
   public void setLockAcquisitionTimeout(long timeout)
   {
      this.lock_acquisition_timeout = timeout;
   }

   /**
    * Max time a lock can be held. After this time the lock is released.
    * (Not currently used.)
    *
    * @return long The max number of milliseconds to hold a lock
    * @jmx.managed-attribute
    */
   public long getLockLeaseTimeout()
   {
      return lock_lease_timeout;
   }

   /**
    * Sets the max time a lock can be held
    *
    * @jmx.managed-attribute
    */
   public void setLockLeaseTimeout(long timeout)
   {
      this.lock_lease_timeout = timeout;
   }

   /**
    * Returns the name of the cache eviction policy (must be an implementation
    * of EvictionPolicy)
    *
    * @return Fully qualified name of a class implementing the EvictionPolicy
    *         interface
    * @jmx.managed-attribute
    */
   public String getEvictionPolicyClass()
   {
      return eviction_policy_class;
   }

   /**
    * Sets the classname of the eviction policy
    *
    * @jmx.managed-attribute
    */
   public void setEvictionPolicyClass(String eviction_policy_class)
   {
      this.eviction_policy_class = eviction_policy_class;
      if (eviction_policy_class == null || eviction_policy_class.length() == 0)
         return;
      try
      {
         eviction_policy_provider_ = (TreeCacheListener)
            getClass().getClassLoader().loadClass(eviction_policy_class).newInstance();
         this.addTreeCacheListener(eviction_policy_provider_);
      }
      catch (Throwable t)
      {
         log.error("setEvictionPolicyClass(): failed creating instance of  " + eviction_policy_class, t);
      }
   }

   /**
    * Obtain eviction thread (if any) wake up interval in seconds
    *
    * @jmx.managed-attribute
    */
   public int getEvictionThreadWakeupIntervalSeconds()
   {
      if (eviction_policy_provider_ == null)
         return -1;
      else
         return ((LRUPolicy) eviction_policy_provider_).getWakeupIntervalSeconds();
   }

   /**
    * Currently not used.
    *
    * @param lm
    */
   public void setLockManager(LockManager lm)
   {
      if (lm != null)
         this.lock_mgr = lm;
   }

   /**
    * Sets the lock manager. The empty constructor will be called. The class
    * needs to implement the <tt>LockManager</tt> interface. Currently not used.
    *
    * @param class_name
    */
   public void setLockManagerClass(String class_name)
   {
      if (class_name == null)
         return;
      try
      {
         LockManager new_lm;
         new_lm = (LockManager) getClass().getClassLoader().loadClass(class_name).newInstance();
         this.lock_mgr = new_lm;
      }
      catch (Throwable t)
      {
         log.error("setLockManager(): failed creating instance of  " + class_name, t);
      }
   }

   /**
    * Sets the TransactionManagerLookup object
    *
    * @param l
    * @jmx.managed-attribute
    */
   public void setTransactionManagerLookup(TransactionManagerLookup l)
   {
      this.tm_lookup = l;
   }


   /**
    * @return
    * @jmx.managed-attribute
    */
   public String getTransactionManagerLookupClass()
   {
      return tm_lookup_class;
   }

   /**
    * Sets the class of the TransactionManagerLookup impl. This will attempt to create an
    * instance, and will throw an exception if this fails.
    *
    * @param cl
    * @throws Exception
    * @jmx.managed-attribute
    */
   public void setTransactionManagerLookupClass(String cl) throws Exception
   {
      this.tm_lookup_class = cl;
   }


   /**
    * Fetch the group state from the current coordinator. If successful,
    * this will trigger setState().
    *
    * @jmx.managed-operation
    */
   public void fetchState(long timeout)
      throws org.jgroups.ChannelClosedException,
      org.jgroups.ChannelNotConnectedException
   {
      boolean rc = channel.getState(null, timeout);
      if (rc)
         log.info("fetchState(): state was retrieved successfully");
      else
         log.info("fetchState(): state could not be retrieved (first member)");
   }

   /**
    * @param listener
    * @jmx.managed-operation
    */
   public void addTreeCacheListener(TreeCacheListener listener)
   {
      if (!listeners.contains(listener))
         listeners.addElement(listener);
   }

   /**
    * @param listener
    * @jmx.managed-operation
    */
   public void removeTreeCacheListener(TreeCacheListener listener)
   {
      listeners.removeElement(listener);
   }

   /* --------------------------- MBeanSupport ------------------------- */

   protected void startService() throws Exception
   {
      if (this.tm_lookup == null && this.tm_lookup_class != null)
      {
         Class clazz = Thread.currentThread().getContextClassLoader().loadClass(this.tm_lookup_class);
         this.tm_lookup = (TransactionManagerLookup) clazz.newInstance();
      }

      try
      {
         if (tm_lookup != null)
            tm = tm_lookup.getTransactionManager();
         else
            log.warn("start(): No transaction manager lookup class has been defined. tx will be null");
      }
      catch (Exception e)
      {
         log.debug("start(): failed looking up TransactionManager, will not use transactions", e);
      }

      if (cache_mode == LOCAL)
      {
         log.info("start(): cache mode is local, will not create the channel");
         notifyCacheStarted();

         return;
      }
      else if (cache_mode == REPL_SYNC)
      {
         log.info("start(): cache mode is repl_sync");
      }
      else if (cache_mode == REPL_ASYNC)
      {
         log.info("start(): cache mode is repl_async");
      }

      if (channel != null)
      { // already started
         log.info("start(): channel is already running");
         notifyCacheStarted();

         return;
      }
      if (cluster_props == null)
      {
         cluster_props = getDefaultProperties();
         log.debug("start(): setting cluster properties to default value");
      }
      channel = new JChannel(cluster_props);
      channel.setOpt(Channel.GET_STATE_EVENTS, new Boolean(true));
      disp = new RpcDispatcher(channel, ml, this, this);
      channel.connect(cluster_name);

      if (fetch_state_on_startup)
      {
         fetchStateOnStartup();
      }

      notifyCacheStarted();

   }

   private void fetchStateOnStartup() throws Exception
   {
      synchronized (stateLock)
      {
         isStateSet = false;
         boolean rc = channel.getState(null, state_fetch_timeout);
         if (rc)
         {
            while (!isStateSet)
            {
               try
               {
                  stateLock.wait();
               }
               catch (InterruptedException iex)
               {
               }
               ;
            }

            log.info("start(): state was retrieved successfully");
         }
         else
            log.info("start(): state could not be retrieved (must be first member in group)");
      }
   }

   protected void stopService()
   {
      if (channel != null)
      {
         log.info("stop(): closing the channel");
         channel.close();
         channel = null;
      }
      if (disp != null)
      {
         log.info("stop(): stopping the dispatcher");
         disp.stop();
         disp = null;
      }
      if (members != null && members.size() > 0)
         members.clear();

      if (repl_queue != null)
         repl_queue.stop();

      notifyCacheStopped();
   }

   /* ----------------------- End of MBeanSupport ----------------------- */

   /**
    * @param fqn fqn String name to retrieve from cache
    * @return Node corresponding to the fqn. Null if does not exist.
    * @jmx.managed-operation
    */
   public Node get(String fqn) throws LockingException, TimeoutException
   {
      return get(Fqn.fromString(fqn));
   }

   /**
    * @param fqn fqn instance to retrieve from cache
    * @return Node corresponding to the fqn. Null if does not exist.
    * @jmx.managed-operation
    */
   public Node get(Fqn fqn) throws LockingException, TimeoutException
   {
      try
      {
         return findNode(fqn, NODE_READ);
      }
      finally
      {
         // release locks since no tx context
         if (locking == true && getCurrentTransaction() == null)
            releaseLocks(null, fqn);
      }
   }


   /**
    * @param fqn
    * @return
    * @jmx.managed-operation
    */
   public Set getKeys(String fqn) throws LockingException, TimeoutException
   {
      return getKeys(Fqn.fromString(fqn));
   }

   /**
    * @param fqn
    * @return
    * @jmx.managed-operation
    */
   public Set getKeys(Fqn fqn) throws LockingException, TimeoutException
   {
      try
      {
         Node n = findNode(fqn, NODE_READ);
         return n != null ? n.getDataKeys() : null;
      }
      finally
      {
         // release locks since no tx context
         if (locking == true && getCurrentTransaction() == null)
            releaseLocks(null, fqn);
      }
   }

   /**
    * Finds a node given its name and returns the value associated with a given key in its <code>data</code>
    * map. Returns null if the node was not found in the tree or the key was not found in the hashmap.
    *
    * @param fqn The fully qualified name of the node.
    * @param key The key.
    * @jmx.managed-operation
    */
   public Object get(String fqn, Object key) throws LockingException, TimeoutException
   {
      return get(Fqn.fromString(fqn), key);
   }


   /**
    * Finds a node given its name and returns the value associated with a given key in its <code>data</code>
    * map. Returns null if the node was not found in the tree or the key was not found in the hashmap.
    *
    * @param fqn The fully qualified name of the node.
    * @param key The key.
    * @jmx.managed-operation
    */
   public Object get(Fqn fqn, Object key) throws LockingException, TimeoutException
   {
      return get(fqn, key, true);
   }

   protected Object get(Fqn fqn, Object key, boolean sendNodeEvent) throws LockingException, TimeoutException
   {
      if (log.isDebugEnabled())
         log.debug("get(" + ", \"" + fqn + "\", " + key + ", \"" + sendNodeEvent + "\")");
      try
      {
         Node n = findNode(fqn, NODE_READ);
         if (n == null) return null;
         if (sendNodeEvent)
            notifyNodeVisisted(fqn);
         return n.get(key);
      }
      finally
      {
         // release locks since no tx context
         if (locking == true && getCurrentTransaction() == null)
            releaseLocks(null, fqn);
      }
   }

   /**
    * Like <code>get</code> method but without triggering a node visit event. This is used
    * to prevent refresh of the cache data in the eviction policy.
    * @param fqn
    * @param key
    * @return
    * @throws LockingException
    * @throws TimeoutException
    */
   public Object peek(Fqn fqn, Object key) throws LockingException, TimeoutException
   {
      return _get(null, fqn, key, false);
   }

   /**
    * Retrieve cache node based on current transaction context.
    *
    * @param gtx existing global transaction context. Must not be null.
    * @param fqn fqn String name to retrieve from cache
    * @param key Key to the obejct value residing in fqn.
    * @return Object corresponding to the value object. Null if does not exist.
    */
   protected Object _get(GlobalTransaction gtx, String fqn, Object key) throws LockingException, TimeoutException
   {
      return _get(gtx, Fqn.fromString(fqn), key);
   }

   /**
    * Retrieve cache node based on current transaction context.
    *
    * @param gtx existing global transaction context. Must not be null.
    * @param fqn fqn instance to retrieve from cache
    * @param key Key to the obejct value residing in fqn.
    * @return Object corresponding to the value object. Null if does not exist.
    */
   protected Object _get(GlobalTransaction gtx, Fqn fqn, Object key) throws LockingException,
      TimeoutException
   {
      boolean sendNodeEvent = false;
      return _get(gtx, fqn, key, sendNodeEvent);
   }

   protected Object _get(GlobalTransaction gtx, Fqn fqn, Object key, boolean sendNodeEvent) throws LockingException,
      TimeoutException
   {
      // No transaction context, need to release the lock afterwards.
      if (gtx == null)
         return get(fqn, key, sendNodeEvent);

      if (log.isDebugEnabled())
         log.debug("_get(" + gtx + ", \"" + fqn + "\", " + key + ")");

      Node n = findNode(fqn, false, locking, gtx, NODE_READ);
      if (n == null) return null;

      // We may not want ot emit event in some cases.
      if (sendNodeEvent)
         notifyNodeVisisted(fqn);

      return n.get(key);
   }

   /**
    * Checks whether a given node exists in the tree
    *
    * @param fqn The fully qualified name of the node
    * @return boolean Whether or not the node exists
    * @jmx.managed-operation
    */
   public boolean exists(String fqn)
   {
      return exists(Fqn.fromString(fqn));
   }

   /**
    * Checks whether a given node exists in the tree
    *
    * @param fqn The fully qualified name of the node
    * @return boolean Whether or not the node exists
    * @jmx.managed-operation
    */
   public boolean exists(Fqn fqn)
   {
      if (fqn == null) return false;
      try
      {
         return findNode(fqn, NODE_READ) != null ? true : false;
      }
      catch (Throwable t)
      {
         log.warn("exists()", t);
         return false;
      }
      finally
      {
         // release locks since no tx context
         if (locking == true && getCurrentTransaction() == null)
            releaseLocks(null, fqn);
      }
   }

   /**
    * Adds a new node to the tree and sets its data. If the node doesn not yet exist, it will be created.
    * Also, parent nodes will be created if not existent. If the node already has data, then the new data
    * will override the old one. If the node already existed, a nodeModified() notification will be generated.
    * Otherwise a nodeAdded() motification will be emitted.
    *
    * @param fqn  The fully qualified name of the new node
    * @param data The new data. May be null if no data should be set in the node.
    * @jmx.managed-operation
    */
   public void put(String fqn, Map data) throws Exception
   {
      put(Fqn.fromString(fqn), data);
   }

   /**
    * Adds a new node to the tree and sets its data. If the node doesn not yet exist, it will be created.
    * Also, parent nodes will be created if not existent. If the node already has data, then the new data
    * will override the old one. If the node already existed, a nodeModified() notification will be generated.
    * Otherwise a nodeAdded() motification will be emitted.
    *
    * @param fqn  The fully qualified name of the new node
    * @param data The new data. May be null if no data should be set in the node.
    * @jmx.managed-operation
    */
   public void put(Fqn fqn, Map data) throws Exception
   {
      GlobalTransaction tx = getCurrentTransaction();
      MethodCall m = new MethodCall(put_data_method,
         new Object[]{tx, fqn, data, new Boolean(true)});
      invokeMethod(tx, m);
   }

   /**
    * Adds a key and value to a given node. If the node doesn't exist, it will be created. If the node
    * already existed, a nodeModified() notification will be generated. Otherwise a
    * nodeAdded() motification will be emitted.
    *
    * @param fqn   The fully qualified name of the node
    * @param key   The key
    * @param value The value
    * @return Object The previous value (if any), if node was present
    * @jmx.managed-operation
    */
   public Object put(String fqn, Object key, Object value) throws Exception
   {
      return put(Fqn.fromString(fqn), key, value);
   }

   /**
    * Adds a key and value to a given node. If the node doesn't exist, it will be created. If the node
    * already existed, a nodeModified() notification will be generated. Otherwise a
    * nodeAdded() motification will be emitted.
    *
    * @param fqn   The fully qualified name of the node
    * @param key   The key
    * @param value The value
    * @return Object The previous value (if any), if node was present
    * @jmx.managed-operation
    */
   public Object put(Fqn fqn, Object key, Object value) throws Exception
   {
      GlobalTransaction tx = getCurrentTransaction();
      MethodCall m = new MethodCall(put_key_val_method,
         new Object[]{tx, fqn, key, value, new Boolean(true)});
      return invokeMethod(tx, m);
   }

   /**
    * Removes the node from the tree.
    *
    * @param fqn The fully qualified name of the node.
    * @jmx.managed-operation
    */
   public void remove(String fqn) throws Exception
   {
      remove(Fqn.fromString(fqn));
   }

   /**
    * Removes the node from the tree.
    *
    * @param fqn The fully qualified name of the node.
    * @jmx.managed-operation
    */
   public void remove(Fqn fqn) throws Exception
   {
      GlobalTransaction tx = getCurrentTransaction();
      MethodCall m = new MethodCall(remove_node_method,
         new Object[]{tx, fqn, new Boolean(true)});
      invokeMethod(tx, m);
   }

   /**
    * Called by eviction policy provider. Note that eviction is done only in local mode,
    * that is, it doesn't replicate the node removal. This is will cause the replcation nodes
    * not synchronizing, but it is ok since user is supposed to add the node again when get is
    * null. After that, the contents will be in sync.
    * @param fqn Will remove everythign assoicated with this fqn.
    * @throws Exception
    */
   public void evict(Fqn fqn) throws Exception
   {
      if (!exists(fqn)) return;   // node does not exist. Maybe it has been recursively removed.
      // use remove method now if there is a child node. Otherwise, it is removed
      boolean create_undo_ops = false;
      boolean sendNodeEvent = false;
      if (hasChild(fqn))
      {
         _removeData(null, fqn, create_undo_ops, sendNodeEvent);
      }
      else
      {
         _remove(null, fqn, create_undo_ops, sendNodeEvent);
      }
   }

   /**
    * Removes <code>key</code> from the node's hashmap
    *
    * @param fqn The fullly qualified name of the node
    * @param key The key to be removed
    * @return The previous value, or null if none was associated with the given key
    * @jmx.managed-operation todo: what about the return value ? e.g. with sync repl we can have many retvals...
    */
   public Object remove(String fqn, Object key) throws Exception
   {
      return remove(Fqn.fromString(fqn), key);
   }

   /**
    * Removes <code>key</code> from the node's hashmap
    *
    * @param fqn The fullly qualified name of the node
    * @param key The key to be removed
    * @return The previous value, or null if none was associated with the given key
    * @jmx.managed-operation todo: what about the return value ? e.g. with sync repl we can have many retvals...
    */
   public Object remove(Fqn fqn, Object key) throws Exception
   {
      GlobalTransaction tx = getCurrentTransaction();
      MethodCall m = new MethodCall(remove_key_method,
         new Object[]{tx, fqn, key, new Boolean(true)});
      return invokeMethod(tx, m);
   }

   public void removeData(String fqn) throws Exception
   {
      removeData(Fqn.fromString(fqn));
   }

   public void removeData(Fqn fqn) throws Exception
   {
      GlobalTransaction tx = getCurrentTransaction();
      MethodCall m = new MethodCall(remove_data_method,
         new Object[]{tx, fqn, new Boolean(true)});
      invokeMethod(tx, m);
   }

   /**
    * Force-releases all locks in this node and the entire subtree
    *
    * @param fqn
    * @jmx.managed-operation
    */
   public void releaseAllLocks(String fqn)
   {
      releaseAllLocks(Fqn.fromString(fqn));
   }

   /**
    * Force-releases all locks in this node and the entire subtree
    *
    * @param fqn
    * @jmx.managed-operation
    */
   public void releaseAllLocks(Fqn fqn)
   {
      Node n;

      try
      {
         n = findNode(fqn, false, false, null, NODE_READ);
         if (n == null)
         {
            log.error("releaseAllLocks(): node " + fqn + " not found");
            return;
         }
         n.releaseAllForce();
      }
      catch (Throwable t)
      {
         log.error("releaseAllLocks(): failed", t);
      }
   }

   /**
    * Prints a representation of the node defined by <code>fqn</code>.
    * Output includes name, fqn and data.
    *
    * @jmx.managed-operation
    */
   public String print(String fqn)
   {
      return print(Fqn.fromString(fqn));
   }

   /**
    * Prints a representation of the node defined by <code>fqn</code>.
    * Output includes name, fqn and data.
    *
    * @jmx.managed-operation
    */
   public String print(Fqn fqn)
   {
      try
      {
         Node n = findNode(fqn, false, false, null, NODE_READ);
         if (n == null) return null;
         return n.toString();
      }
      catch (Throwable t)
      {
         return null;
      }
   }


   /**
    * Returns all children of a given node
    *
    * @param fqn The fully qualified name of the node
    * @return Set A list of child names (as Strings)
    * @jmx.managed-operation
    */
   public Set getChildrenNames(String fqn) throws LockingException, TimeoutException
   {
      return getChildrenNames(Fqn.fromString(fqn));
   }

   /**
    * Returns all children of a given node
    *
    * @param fqn The fully qualified name of the node
    * @return Set A list of child names (as Strings)
    * @jmx.managed-operation
    */
   public Set getChildrenNames(Fqn fqn) throws LockingException, TimeoutException
   {
      try
      {
         Node n = findNode(fqn, NODE_READ);
         if (n == null) return null;
         Map m = n.getChildren();
         if (m != null)
            return m.keySet();
         else
            return null;
      }
      finally
      {
         if (locking == true && getCurrentTransaction() == null)
         {
            releaseLocks(null, fqn);
         }
      }
   }

   public boolean hasChild(Fqn fqn)
   {
      try
      {
         Node n = findNode(fqn, false, false, null, NODE_READ);
//         if(n == null) return false;
         return n.hasChild();
      }
      catch (LockingException e)
      {
         e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
      }
      catch (TimeoutException e)
      {
         e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
      }
      return false;
   }

   /**
    * @return
    * @jmx.managed-operation
    */
   public String toString()
   {
      StringBuffer sb = new StringBuffer();
      int indent = 0;
      Map children;

      children = root.getChildren();
      if (children != null && children.size() > 0)
      {
         Collection nodes = children.values();
         for (Iterator it = nodes.iterator(); it.hasNext();)
         {
            ((Node) it.next()).print(sb, indent);
            sb.append("\n");
         }
      }
      else
         sb.append(SEPARATOR);
      return sb.toString();
   }

   /**
    * @return
    * @jmx.managed-operation
    */
   public String printDetails()
   {
      StringBuffer sb = new StringBuffer();
      int indent = 0;
      Map children;

      children = root.getChildren();
      if (children != null && children.size() > 0)
      {
         Collection nodes = children.values();
         for (Iterator it = nodes.iterator(); it.hasNext();)
         {
            ((Node) it.next()).printDetails(sb, indent);
            sb.append("\n");
         }
      }
      else
         sb.append(SEPARATOR);
      return sb.toString();
   }

   /**
    * @return
    * @jmx.managed-operation
    */
   public String printLockInfo()
   {
      StringBuffer sb = new StringBuffer();
      int indent = 0;
      Map children;

      children = root.getChildren();
      if (children != null && children.size() > 0)
      {
         Collection nodes = children.values();
         for (Iterator it = nodes.iterator(); it.hasNext();)
         {
            ((Node) it.next()).printLockInfo(sb, indent);
            sb.append("\n");
         }
      }
      else
         sb.append(SEPARATOR);
      return sb.toString();
   }

   /**
    * Prints the entire configuration (all variables). Uses reflection to do so
    */
   public String printInfo()
   {
      StringBuffer sb = new StringBuffer();
      Class cl = getClass();
      Field[] fields = cl.getDeclaredFields();

      for (int i = 0; i < fields.length; i++)
      {
         Field f = fields[i];
         sb.append(f.getName()).append(": ");
         try
         {
            sb.append(f.get(this));
         }
         catch (Exception e)
         {
            e.printStackTrace();
         }
         sb.append("\n");
      }

      return sb.toString();
   }

   /* ---------------------- Remote method calls -------------------- */

   public List callRemoteMethods(Vector mbrs, MethodCall method_call,
      boolean synchronous, boolean exclude_self, long timeout)
      throws Exception
   {
      RspList rsps;
      Rsp rsp;
      List retval;
      Vector validMembers;
      int mode = synchronous ? GroupRequest.GET_ALL : GroupRequest.GET_NONE;

      if (disp == null)
         return null;

      validMembers = mbrs != null ? new Vector(mbrs) : new Vector(this.members);
      if (exclude_self && validMembers != null && validMembers.size() > 0)
      {
         Object local_addr = getLocalAddress();
         if (local_addr != null)
            validMembers.remove(local_addr);
      }
      if (validMembers.size() == 0)
      {
         if (log.isDebugEnabled())
            log.debug("callRemoteMethods(): destination list is empty, discarding call");
         return null;
      }

      if (log.isDebugEnabled())
         log.debug("callRemoteMethods(): valid members are " + validMembers);

      rsps = disp.callRemoteMethods(validMembers, method_call, mode, timeout);
      if (rsps == null)
         return null;
      retval = new ArrayList(rsps.size());
      for (int i = 0; i < rsps.size(); i++)
      {
         rsp = (Rsp) rsps.elementAt(i);
         retval.add(rsp.getValue());
      }
      return retval;
   }


   /**
    * @param members
    * @param method
    * @param args
    * @param synchronous
    * @param exclude_self
    * @param timeout
    * @return
    * @throws Exception
    * @jmx.managed-operation
    */
   public List callRemoteMethods(Vector members, Method method, Object[] args,
      boolean synchronous, boolean exclude_self, long timeout)
      throws Exception
   {
      return callRemoteMethods(members, new MethodCall(method, args), synchronous,
         exclude_self, timeout);
   }

   /**
    * @param members
    * @param method_name
    * @param types
    * @param args
    * @param synchronous
    * @param exclude_self
    * @param timeout
    * @return
    * @throws Exception
    * @jmx.managed-operation
    */
   public List callRemoteMethods(Vector members, String method_name,
      Class[] types, Object[] args,
      boolean synchronous, boolean exclude_self, long timeout)
      throws Exception
   {
      Method method = getClass().getMethod(method_name, types);
      return callRemoteMethods(members, method, args, synchronous, exclude_self, timeout);
   }
   /* -------------------- End Remote method calls ------------------ */


   /* --------------------- Callbacks -------------------------- */

   /**
    * Does the real work. Needs to acquire locks if accessing nodes, depending on
    * the value of <tt>locking</tt>. If run inside a transaction, needs to (a) add
    * newly acquired locks to {@link TransactionEntry}'s lock list, (b) add nodes
    * that were created to {@link TransactionEntry}'s node list and (c) create
    * {@link Modification}s and add them to {@link TransactionEntry}'s modification
    * list and (d) create compensating modifications to undo the changes in case
    * of a rollback
    *
    * @param fqn
    * @param data
    * @param create_undo_ops If true, undo operations will be created (default is true).
    *                        Otherwise they will not be created (used by rollback()).
    * @jmx.managed-operation
    */
   public void _put(GlobalTransaction tx, String fqn, Map data, boolean create_undo_ops)
      throws LockingException, TimeoutException
   {
      _put(tx, Fqn.fromString(fqn), data, create_undo_ops);
   }


   /**
    * Does the real work. Needs to acquire locks if accessing nodes, depending on
    * the value of <tt>locking</tt>. If run inside a transaction, needs to (a) add
    * newly acquired locks to {@link TransactionEntry}'s lock list, (b) add nodes
    * that were created to {@link TransactionEntry}'s node list and (c) create
    * {@link Modification}s and add them to {@link TransactionEntry}'s modification
    * list and (d) create compensating modifications to undo the changes in case
    * of a rollback
    *
    * @param fqn
    * @param data
    * @param create_undo_ops If true, undo operations will be created (default is true).
    *                        Otherwise they will not be created (used by rollback()).
    * @jmx.managed-operation
    */
   public void _put(GlobalTransaction tx, Fqn fqn, Map data, boolean create_undo_ops)
      throws LockingException, TimeoutException
   {
      _put(tx, fqn, data, create_undo_ops, false);
   }


   /**
    * Does the real work. Needs to acquire locks if accessing nodes, depending on
    * the value of <tt>locking</tt>. If run inside a transaction, needs to (a) add
    * newly acquired locks to {@link TransactionEntry}'s lock list, (b) add nodes
    * that were created to {@link TransactionEntry}'s node list and (c) create
    * {@link Modification}s and add them to {@link TransactionEntry}'s modification
    * list and (d) create compensating modifications to undo the changes in case
    * of a rollback
    *
    * @param fqn
    * @param data
    * @param create_undo_ops If true, undo operations will be created (default is true).
    * @param erase_contents  Clear the existing hashmap before putting the new data into it
    *                        Otherwise they will not be created (used by rollback()).
    * @jmx.managed-operation
    */
   public void _put(GlobalTransaction tx, Fqn fqn, Map data, boolean create_undo_ops, boolean erase_contents)
      throws LockingException, TimeoutException
   {
      Node n;
      MethodCall undo_op = null;
      Map old_data;

      if (log.isDebugEnabled())
         log.debug("_put(" + tx + ", \"" + fqn + "\", " + data + ")");

      try
      {
         // Find the node. This will lock it (if <tt>locking</tt> is true) and
         // add the temporarily created parent nodes to the TX's node list if tx != null)
         n = findNode(fqn, true, locking, tx, NODE_WRITE);
         if (n == null)
         {
            log.error("_put(): node " + fqn + " not found");
            return;
         }

         // create a compensating method call (reverting the effect of
         // this modification) and put it into the TX's undo list.
         if (tx != null && create_undo_ops)
         {
            if ((old_data = n.getData()) == null)
            {
               undo_op = new MethodCall(remove_data_method,
                  new Object[]{tx, fqn, new Boolean(false)});
            }
            else
            {
               undo_op = new MethodCall(put_data_erase_method,
                  new Object[]{tx, fqn,
                               new HashMap(old_data),
                               new Boolean(false),
                               new Boolean(true)}); //erase previous hashmap contents
            }
         }

         if (erase_contents)
            n.put(data, true);
         else
            n.put(data);

         notifyNodeModified(fqn);

         if (tx != null && create_undo_ops)
         {
            // 1. put undo-op in TX' undo-operations list (needed to rollback TX)
            tx_table.addUndoOperation(tx, undo_op);
         }
      }
      finally
      { // we always unlock when no TX is present (lock-update-unlock)
         if (tx == null && locking == true)
            releaseLocks(null, fqn);
      }
   }


   /**
    * @param fqn
    * @param key
    * @param value
    * @return Previous value (if any)
    * @jmx.managed-operation
    */
   public Object _put(GlobalTransaction tx, String fqn, Object key, Object value,
      boolean create_undo_ops)
      throws LockingException, TimeoutException
   {
      return _put(tx, Fqn.fromString(fqn), key, value, create_undo_ops);
   }


   /**
    * @param fqn
    * @param key
    * @param value
    * @return Previous value (if any)
    * @jmx.managed-operation
    */
   public Object _put(GlobalTransaction tx, Fqn fqn, Object key, Object value,
      boolean create_undo_ops)
      throws LockingException, TimeoutException
   {
      Node n = null;
      MethodCall undo_op = null;
      Object old_value = null;

      if (log.isDebugEnabled())
         log.debug("_put(" + tx + ", \"" + fqn + "\", " + key + ", " + value + ")");

      try
      {
         // Find the node. This will lock it (if <tt>locking</tt> is true) and
         // add the temporarily created parent nodes to the TX's node list if tx != null)
         n = findNode(fqn, true, locking, tx, NODE_WRITE);
         if (n == null)
         {
            log.error("_put(): node " + fqn + " not found");
            return null;
         }
         old_value = n.put(key, value);

         // create a compensating method call (reverting the effect of
         // this modification) and put it into the TX's undo list.
         if (tx != null && create_undo_ops)
         {
            if (old_value == null)
            {
               undo_op = new MethodCall(remove_key_method,
                  new Object[]{tx, fqn, key, new Boolean(false)});
            }
            else
            {
               undo_op = new MethodCall(put_key_val_method,
                  new Object[]{tx, fqn, key, old_value,
                               new Boolean(false)});
            }
            // 1. put undo-op in TX' undo-operations list (needed to rollback TX)
            tx_table.addUndoOperation(tx, undo_op);
         }
         notifyNodeModified(fqn);
         return old_value;
      }
      finally
      { // we always unlock when no TX is present (lock-update-unlock)
         if (tx == null && locking == true)
            releaseLocks(null, fqn);
      }
   }

   /**
    * @param fqn todo: check if removal of root node ("/") is handled correctly
    * @jmx.managed-operation
    */
   public void _remove(GlobalTransaction tx, String fqn, boolean create_undo_ops)
      throws LockingException, TimeoutException
   {
      _remove(tx, Fqn.fromString(fqn), create_undo_ops);
   }

   /**
    * @param fqn todo: check if removal of root node ("/") is handled correctly
    * @jmx.managed-operation
    */
   public void _remove(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops)
      throws LockingException, TimeoutException
   {
      _remove(tx, fqn, create_undo_ops, true);
   }

   public void _remove(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops, boolean sendNodeEvent)
      throws LockingException, TimeoutException
   {
      Node n, parent_node;
      MethodCall undo_op = null;

      if (log.isDebugEnabled())
         log.debug("_remove(" + tx + ", \"" + fqn + "\")");

      if (fqn.size() == 0)
      {
         log.error("_remove(): removal of root node is not yet implemented. " +
            "To do this, remove all child nodes of root separately");
         return;
      }

      try
      {
         // Find the node. This will lock it (if <tt>locking</tt> is true) and
         // add the temporarily created parent nodes to the TX's node list if tx != null)
         n = findNode(fqn, true, locking, tx, NODE_WRITE);
         if (n == null)
         {
            log.error("_remove(): node " + fqn + " not found");
            return;
         }
         parent_node = n.getParent();

         // release all locks for n
         if (locking)
         {
            if (tx != null)
               n.releaseAll(tx);
            else
               n.releaseAll(getGlobalTransactionFromThread());
         }

         // remove subtree from parent
         parent_node.removeChild(n.getName());

         // create a compensating method call (reverting the effect of
         // this modification) and put it into the TX's undo list.
         if (tx != null && create_undo_ops && n != null)
         {
            undo_op = new MethodCall(addChild_method,
               new Object[]{tx, parent_node.getFqn(),
                            n.getName(), n});

            // 1. put undo-op in TX' undo-operations list (needed to rollback TX)
            tx_table.addUndoOperation(tx, undo_op);
         }

         if (sendNodeEvent)
            notifyNodeRemoved(fqn);
      }
      finally
      { // we always unlock when no TX is present (lock-update-unlock)
         if (tx == null && locking == true)
            releaseLocks(null, fqn);
      }
   }

   /**
    * @param fqn
    * @param key
    * @return Object
    * @jmx.managed-operation
    */
   public Object _remove(GlobalTransaction tx, String fqn, Object key, boolean create_undo_ops)
      throws LockingException, TimeoutException
   {
      return _remove(tx, Fqn.fromString(fqn), key, create_undo_ops);
   }

   /**
    * @param fqn
    * @param key
    * @return Object
    * @jmx.managed-operation
    */
   public Object _remove(GlobalTransaction tx, Fqn fqn, Object key, boolean create_undo_ops)
      throws LockingException, TimeoutException
   {
      return _remove(tx, fqn, key, create_undo_ops, false); // din't node remove notification event
   }

   public Object _remove(GlobalTransaction tx, Fqn fqn, Object key, boolean create_undo_ops, boolean sendNodeEvent)
      throws LockingException, TimeoutException
   {
      Node n = null;
      MethodCall undo_op = null;
      Object old_value = null;

      if (log.isDebugEnabled())
         log.debug("_remove(" + tx + ", \"" + fqn + "\", " + key + ")");

      try
      {
         // Find the node. This will lock it (if <tt>locking</tt> is true) and
         // add the temporarily created parent nodes to the TX's node list if tx != null)
         n = findNode(fqn, true, locking, tx, NODE_WRITE);
         if (n == null)
         {
            log.error("_remove(): node " + fqn + " not found");
            return null;
         }
         old_value = n.remove(key);

         // create a compensating method call (reverting the effect of
         // this modification) and put it into the TX's undo list.
         if (tx != null && create_undo_ops && old_value != null)
         {
            undo_op = new MethodCall(put_key_val_method,
               new Object[]{tx, fqn, key, old_value,
                            new Boolean(false)});
            // 1. put undo-op in TX' undo-operations list (needed to rollback TX)
            tx_table.addUndoOperation(tx, undo_op);
         }

         if (sendNodeEvent)
            notifyNodeRemoved(fqn);
         return old_value;
      }
      finally
      { // we always unlock when no TX is present (lock-update-unlock)
         if (tx == null && locking == true)
            releaseLocks(tx, fqn);
      }
   }

   /**
    * @param fqn
    */
   public void _removeData(GlobalTransaction tx, String fqn, boolean create_undo_ops)
      throws LockingException, TimeoutException
   {
      _removeData(tx, Fqn.fromString(fqn), create_undo_ops);
   }

   /**
    * @param fqn
    */
   public void _removeData(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops)
      throws LockingException, TimeoutException
   {
      _removeData(tx, fqn, create_undo_ops, true);
   }

   public void _removeData(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops, boolean sendNodeEvent)
      throws LockingException, TimeoutException
   {
      Node n = null;
      MethodCall undo_op = null;
      Map old_data = null;

      if (log.isDebugEnabled())
         log.debug("_removeData(" + tx + ", \"" + fqn + "\")");

      try
      {
         // Find the node. This will lock it (if <tt>locking</tt> is true) and
         // add the temporarily created parent nodes to the TX's node list if tx != null)
         n = findNode(fqn, true, locking, tx, NODE_WRITE);
         if (n == null)
         {
            log.error("_removeData(): node " + fqn + " not found");
            return;
         }

         // create a compensating method call (reverting the effect of
         // this modification) and put it into the TX's undo list.
         if (tx != null && create_undo_ops && (old_data = n.getData()) != null)
         {
            undo_op = new MethodCall(put_data_method,
               new Object[]{tx, fqn,
                            new HashMap(old_data),
                            new Boolean(false)});
         }

         n.clear();

         if (sendNodeEvent)
            notifyNodeVisisted(fqn);
         // put undo-op in TX' undo-operations list (needed to rollback TX)
         if (tx != null && create_undo_ops)
         {
            tx_table.addUndoOperation(tx, undo_op);
         }
      }
      finally
      { // we always unlock when no TX is present (lock-update-unlock)
         if (tx == null && locking == true)
            releaseLocks(null, fqn);
      }
   }

   /**
    * Compensating method to {@link #_remove(GlobalTransaction,Fqn,boolean)}. This
    * is package-private on purpose because it is only supposed to be called inside
    * the same VM (by {@link #rollback(GlobalTransaction)}).
    *
    * @param parent_fqn
    * @param child_name
    * @param old_node
    */
   void _addChild(GlobalTransaction tx, String parent_fqn, Object child_name, Node old_node)
      throws LockingException, TimeoutException
   {
      _addChild(tx, Fqn.fromString(parent_fqn), child_name, old_node);
   }

   /**
    * Compensating method to {@link #_remove(GlobalTransaction,Fqn,boolean)}. This
    * is package-private on purpose because it is only supposed to be called inside
    * the same VM (by {@link #rollback(GlobalTransaction)}).
    *
    * @param parent_fqn
    * @param child_name
    * @param old_node
    */
   public void _addChild(GlobalTransaction tx, Fqn parent_fqn, Object child_name, Node old_node)
      throws LockingException, TimeoutException
   {
      if (log.isDebugEnabled())
         log.debug("_addChild(" + tx + ", \"" + parent_fqn + "\", \"" + child_name +
            "\")");

      if (parent_fqn == null || child_name == null || old_node == null)
      {
         log.error("_addChild(): parent_fqn or child_name or node was null");
         return;
      }
      Node tmp = findNode(parent_fqn, false, false, tx, NODE_WRITE);
      if (tmp == null)
      {
         log.error("_addChild(): node " + parent_fqn + " not found");
         return;
      }

// TODO This is bit weird????
      if (locking)
      {
         GlobalTransaction gtx = null;
         if (tx != null)
            gtx = tx;
         else
            gtx = getGlobalTransactionFromThread();

         old_node.releaseAll(gtx);
         old_node.acquireAll(gtx, lock_acquisition_timeout, Node.LOCK_TYPE_WRITE);
      }
      tmp.addChild(child_name, old_node);
   }


   /* ----------------- End of  Callbacks ---------------------- */


   /* ----------------- 2-phase commit protocol----------------- */


   /**
    * Calls {@link #prepare(GlobalTransaction,List,org.jgroups.Address,boolean))} in all members except self.
    * Waits for all responses. If one of the members failed to prepare, its return value
    * will be an exception. If there is one exception we rethrow it. This will mark the
    * current transaction as rolled back, which will cause the
    * {@link SynchronizationHandler#afterCompletion(int)} callback to have a status
    * of <tt>MARKED_ROLLBACK</tt>. When we get that call, we simply roll back the
    * transaction.<br/>
    * If everything runs okay, the {@link SynchronizationHandler#afterCompletion(int)}
    * callback will trigger the {@link #runCommitPhase(GlobalTransaction))}.
    * <br/>
    *
    * @param tx
    * @param coordinator
    * @param async       Run this asynchronously
    * @param commit      Commit immediately after prepare() is successful ?
    * @throws Exception
    */
   protected void runPreparePhase(GlobalTransaction tx,
      Address coordinator,
      boolean async,
      boolean commit) throws Exception
   {
      List rsps, modifications;
      TransactionEntry entry;

      entry = tx_table.get(tx);
      if (entry == null)
         throw new Exception("runPreparePhase(): found no entry for transaction " + tx);
      modifications = entry.getModifications();

      // this method will return immediately if we're the only member (because exclude_self=true)
      if (log.isDebugEnabled())
         log.debug("runPreparePhase(): run remote prepare with async mode: " + async +
            " commit: " + commit + " and coordinator: " + coordinator);

      rsps = callRemoteMethods(members, prepare_method,
         new Object[]{tx, modifications, (Address) getLocalAddress(),
                      new Boolean(commit)},
         !async, // sync or async call ?
         true, // exclude self
         sync_repl_timeout);
      if (rsps != null)
         checkResponses(rsps); // throws an exception if one of the rsps is an exception
   }

   /**
    * Asynchronously calls {@link #commit(GlobalTransaction)} in all members
    *
    * @param tx
    */
   protected void runCommitPhase(GlobalTransaction tx)
   {
      int num_modifications = numberOfModifications(tx);

      // 1. Multicast commit() to all members (exclude myself though)
      //    Only do this if we previously replicated modifications (PREPARE phase)
      if (num_modifications > 0)
      {
         try
         {
            callRemoteMethods(members, commit_method,
               new Object[]{tx},
               false, // this is an async call
               true, // exclude self
               sync_repl_timeout);
         }
         catch (Throwable e)
         {
            log.error("runCommitPhase(): failed", e);
         }
      }

      // 2. Commit locally
      try
      {
         commit(tx);
      }
      catch (Exception e)
      {
         log.error("runCommitPhase(): failed", e);
      }
   }

   /**
    * Asynchronously calls {@link #rollback(GlobalTransaction)} in all members
    *
    * @param tx
    */
   protected void runRollbackPhase(GlobalTransaction tx)
   {
      int num_modifications = numberOfModifications(tx);

      // 1. Multicast rollback() to all other members (excluding myself)
      //    Only do this if we previously replicated modifications (PREPARE phase)
      if (num_modifications > 0)
      {
         try
         {
            callRemoteMethods(members, rollback_method,
               new Object[]{tx},
               false, // this is an async call
               true, // exclude self
               sync_repl_timeout);
         }
         catch (Throwable e)
         {
            log.error("runRollbackPhase(): failed", e);
         }
      }

      // 2. Rollback locally
      try
      {
         rollback(tx);
      }
      catch (Exception e)
      {
         log.error("runRollbackPhase(): failed", e);
      }
   }

   /**
    * @param tx            The global transaction under which all updates are coordinated across
    *                      a cluster
    * @param modifications A list of {@link MethodCall} objects
    * @param coordinator   The address of the member who coordinates this 2pc protocol
    * @param commit        Commit or rollback immediately ?
    * @throws Exception
    */
   public void prepare(GlobalTransaction tx, List modifications, Address coordinator,
      boolean commit) throws Exception
   {
      MethodCall method_call;
      Object retval;
      boolean success = true;

      // discard own message. should never happen because runPreparePhase() calls
      // every member except self !
      if (coordinator != null && coordinator.equals(getLocalAddress()))
      {
         log.warn("prepare(): received my own message (will discard it)");
         return;
      }

      // 1. Create new entry for TX in tx_table, the modifications below will need this entry to add
      //    their modifications under the GlobalTx key
      if (tx_table.get(tx) == null)
      {
         TransactionEntry ent = new TransactionEntry();

         // we don't need the local TX in prepare/commit (bela Nov 13 2003)
         // ent.setTransaction(getLocalTransaction());

         tx_table.put(tx, ent);
      }

      if (log.isDebugEnabled())
         log.debug("prepare(): run prepare ");
      // 2. Run all MethodCalls in modification list
      try
      {
         if (modifications != null)
         {
            for (Iterator it = modifications.iterator(); it.hasNext();)
            {
               method_call = (MethodCall) it.next();
               try {
                  retval = method_call.invoke(this, null);
               }
               catch (Throwable t) {
                  retval=t;
               }
               if (retval != null && retval instanceof Exception)
               {
                  success = false;
                  throw (Exception) retval;
               }
               else
               {
                  // add the modification to the TX's modification list.
                  tx_table.addModification(tx, method_call);
               }
            }
         }
      }
      finally
      {
         if (commit)
         {
            if (success)
               commit(tx);
            else
               rollback(tx);
         }
      }
   }

   /**
    * Remove all locks held by <tt>tx</tt>, remove the transaction from the
    * transaction table
    *
    * @param tx
    */
   public void commit(GlobalTransaction tx)
   {
      if (log.isDebugEnabled())
         log.debug("commit(): called to commit cache with gtx " + tx);

      TransactionEntry entry = tx_table.get(tx);
      if (entry == null)
      {
         log.error("commit(): entry for transaction " + tx + " not found");
         return;
      }

      // Let's do it in stack style, LIFO
      List list = entry.getLocks();
      for (int i = list.size() - 1; i >= 0; i--)
      {
         IdentityLock lock = (IdentityLock) list.get(i);
         if (log.isDebugEnabled())
            log.debug("commit(): releasing lock " + lock);
         lock.release(tx);
      }

      if (log.isDebugEnabled())
         log.debug("commit(): remove local transaction key " + entry.getTransaction());
      tx_table.remove(entry.getTransaction());
      tx_table.remove(tx);
   }

   /**
    * Revert all changes made inside this TX: invoke all method calls of the undo-ops
    * list. Then release all locks and remove the TX from the transaction table.
    * <ol>
    * <li>Revert all modifications done in the current TX<li/>
    * <li>Release all locks held by the current TX</li>
    * <li>Remove all temporary nodes created by the current TX</li>
    * </ol>
    *
    * @param tx
    */
   public void rollback(GlobalTransaction tx)
   {
      List undo_ops;
      TransactionEntry entry = tx_table.get(tx);
      MethodCall undo_op;
      Object retval;
      Fqn node_name;

      if (log.isDebugEnabled())
         log.debug("rollback(): called to rollback cache");

      if (entry == null)
      {
         log.error("rollback(): entry for transaction " + tx +
            " not found (transaction  has possibly already been rolled back)");
         return;
      }

      // 1. Revert the modifications by running the undo-op list in reverse. This
      //    *cannot* throw any exceptions !
      undo_ops = entry.getUndoOperations();
      for (ListIterator it = undo_ops.listIterator(undo_ops.size()); it.hasPrevious();)
      {
         undo_op = (MethodCall) it.previous();
         try
         {
            retval = undo_op.invoke(this, null);
            if (retval != null && retval instanceof Throwable)
            {
               log.error("rollback(): undo operation failed, error=" + retval);
            }
         }
         catch (Throwable t)
         {
            log.error("rollback(): undo operation failed", t);
         }
      }

      // 2. Remove all temporary nodes. Need to do it backward since node is LIFO.
      for (ListIterator it = entry.getNodes().listIterator(entry.getNodes().size());
           it.hasPrevious();)
      {
         node_name = (Fqn) it.previous();
         try
         {
            _remove(tx, node_name, false);
         }
         catch (Throwable t)
         {
            log.error("rollback(): failed removing node \"" + node_name + "\"", t);
         }
      }


      // 3. Finally, release all locks held by this TX
      // Let's do it in stack style, LIFO
      // Note that the lock could have been released already so don't panic.
      List list = entry.getLocks();
      for (int i = list.size() - 1; i >= 0; i--)
      {
         IdentityLock lock = (IdentityLock) list.get(i);
         if (log.isDebugEnabled())
            log.debug("rollback(): releasing lock " + lock);
         lock.release(tx);
      }

      tx_table.remove(entry.getTransaction());
      tx_table.remove(tx);
   }

   /* -------------- End of 2-phase commit protocol------------- */


   /*-------------------- MessageListener ----------------------*/

   class MessageListenerAdaptor implements MessageListener
   {

      /**
       * Callback (no-op)
       *
       * @param msg
       */
      public void receive(Message msg)
      {
         ;
      }

      /**
       * Return a copy of the current cache (tree)
       */
      public byte[] getState()
      {
         GlobalTransaction tx;

         tx = getCurrentTransaction();
         if (tx == null) tx = getGlobalTransactionFromThread();

         try
         {
            log.info("getState(): locking tree");
            root.acquireAll(tx, state_fetch_timeout, Node.LOCK_TYPE_READ);
            log.info("getState(): returning the current state");
            return Util.objectToByteBuffer(root);
         }
         catch (Throwable t)
         {
            log.error("getState(): failed locking tree", t);
            return null;
         }
         finally
         {
            root.releaseAll(tx);
         }
      }

      /**
       * Set the cache (tree) to this value
       */
      public void setState(byte[] new_state)
      {
         Node new_root = null, old_root = null;
         Object obj;
         GlobalTransaction tx;

         if (new_state == null)
         {
            log.info("setState(): new cache is null (maybe first member in cluster)");
            return;
         }
         try
         {
            tx = getCurrentTransaction();
            if (tx == null) tx = getGlobalTransactionFromThread();

            log.info("setState(): setting the new state");
            obj = Util.objectFromByteBuffer(new_state);
            new_root = (Node) obj;

            try
            {
               log.info("setState(): locking the old tree");
               root.acquireAll(tx, state_fetch_timeout, Node.LOCK_TYPE_WRITE);
               log.info("setState(): locking the old tree was successful");
               old_root = root;
               root = new_root;
               notifyAllNodesCreated(root);
            }
            catch (Throwable t)
            {
               log.error("setState(): locking/unlocking the old tree failed", t);
            }
            finally
            {
               if (old_root != null)
               {
                  log.info("setState(): forcing release of all locks in old tree");
                  old_root.releaseAllForce();
               }
            }
         }
         catch (Throwable ex)
         {
            log.error("setState(): could not set cache: ", ex);
         }

         synchronized (stateLock)
         {
            // Notify wait that state has been set.
            isStateSet = true;
            stateLock.notifyAll();
         }
      }

   }

   /*-------------------- End of MessageListener ----------------------*/


   /*----------------------- MembershipListener ------------------------*/

   public void viewAccepted(View new_view)
   {
      Vector new_mbrs = new_view.getMembers();

      // todo: if MergeView, fetch and reconcile state from coordinator
      // actually maybe this is best left up to the application ? we just notify them and let
      // the appl handle it ?

      log.info("viewAccepted(): new members: " + new_mbrs);
      if (new_mbrs != null)
      {
         members.removeAllElements();
         members.addAll(new_view.getMembers());
         notifyViewChange(new_view);
      }
   }


   /**
    * Called when a member is suspected
    */
   public void suspect(Address suspected_mbr)
   {
      ;
   }


   /**
    * Block sending and receiving of messages until viewAccepted() is called
    */
   public void block()
   {
   }

   /*------------------- End of MembershipListener ----------------------*/


   /* ------------------------------ Private methods --------------------------- */

   /**
    * Returns the transaction associated with the current thread. We get the initial context
    * and a reference to the TransactionManager to get the transaction. This method is used
    * by {@link #getCurrentTransaction()}
    *
    * @return
    */
   protected Transaction getLocalTransaction()
   {
      if (tm == null)
      {
         return null;
      }
      try
      {
         return tm.getTransaction();
      }
      catch (Throwable t)
      {
         return null;
      }
   }

   /**
    * Get the transaction associated with the current thread and maps it onto a
    * {@link GlobalTransaction}. The mapping is maintained in the <tt>tx_map</tt> table.
    * If a local transaction exists, but doesn't yet have a mapping to a GlobalTransaction,
    * a new GlobalTransaction will be created and mapped to the local transaction.
    *
    * @return A GlobalTransaction, or null if no (local) transaction was associated with
    *         the current thread
    */
   protected GlobalTransaction getCurrentTransaction()
   {
      GlobalTransaction gtx = null;
      Transaction tx;
      Address addr = (Address) getLocalAddress();

      if ((tx = getLocalTransaction()) == null)
      { // no transaction is associated with the current thread
         return null;
      }

      gtx = tx_table.get(tx);
      if (gtx == null)
      {
         gtx = GlobalTransaction.create(addr);
         tx_table.put(tx, gtx);
         TransactionEntry ent = new TransactionEntry();
         ent.setTransaction(tx);
         tx_table.put(gtx, ent);
         try
         {
            tx.registerSynchronization(new SynchronizationHandler(gtx, this));
            if (log.isDebugEnabled())
               log.debug("getCurrentTransaction(): registered with transaction " +
                  tx + " as Synchronization listener with global transaction: " +
                  gtx);
         }
         catch (Throwable t)
         {
            log.error("getCurrentTransaction(): registration with transaction failed", t);
         }
      }
      return gtx;
   }

   protected void initMethods() throws NoSuchMethodException
   {
      put_data_method = getClass().getMethod("_put",
         new Class[]{GlobalTransaction.class,
                     Fqn.class,
                     Map.class,
                     boolean.class});
      put_data_erase_method = getClass().getMethod("_put",
         new Class[]{GlobalTransaction.class,
                     Fqn.class,
                     Map.class,
                     boolean.class,
                     boolean.class});
      put_key_val_method = getClass().getMethod("_put",
         new Class[]{GlobalTransaction.class,
                     Fqn.class,
                     Object.class,
                     Object.class,
                     boolean.class});
      remove_node_method = getClass().getMethod("_remove",
         new Class[]{GlobalTransaction.class,
                     Fqn.class,
                     boolean.class});
      remove_key_method = getClass().getMethod("_remove",
         new Class[]{GlobalTransaction.class,
                     Fqn.class,
                     Object.class,
                     boolean.class});
      remove_data_method = getClass().getMethod("_removeData",
         new Class[]{GlobalTransaction.class,
                     Fqn.class,
                     boolean.class});
      prepare_method = getClass().getMethod("prepare",
         new Class[]{GlobalTransaction.class,
                     List.class,
                     Address.class,
                     boolean.class});
      commit_method = getClass().getMethod("commit",
         new Class[]{GlobalTransaction.class});
      rollback_method = getClass().getMethod("rollback",
         new Class[]{GlobalTransaction.class});
      addChild_method = getClass().getMethod("_addChild",
         new Class[]{GlobalTransaction.class,
                     Fqn.class, Object.class, Node.class});
   }

   /**
    * Invokes a method against this object. Contains the logger_ic for handling the various
    * use cases, e.g. mode (local, repl_async, repl_sync), transaction (yes or no) and
    * locking (yes or no).
    *
    * @param tx
    * @param m
    * @return
    * @throws Exception
    */
   protected Object invokeMethod(GlobalTransaction tx, MethodCall m) throws Exception
   {
      Object retval = null;
      if (m == null)
      {
         log.error("invokeMethod(): method is null");
         return null;
      }

      if (cache_mode == LOCAL)
      {
         if (log.isDebugEnabled())
            log.debug("invokeMethod(): calling method " + m.getName());
         try {
            retval = m.invoke(this, null);
         }
         catch (Throwable t) {
            retval=t;
         }
         if (retval instanceof RuntimeException)
            throw (RuntimeException) retval;
         else if (retval instanceof Exception)
            throw (Exception) retval;
         else
            return retval;
      }

      if (tx != null)
      {
         return invokeTransactionalMethod(tx, m);
      }
      else
      {
         return invokeTransactionlessMethod(m);
      }
   }

   /**
    * Invokes a method within a transaction. If the update is successful, the modification
    * will be stored in the TX's modification list. Note that the <em>undo-operation</em>
    * will be created in the actual operation (e.g. <tt>_put()</tt>) and added to the
    * TX's undo-operation list (needed for rollback).
    *
    * @param tx The transaction. Guaranteed not to be null.
    * @param m
    * @return
    */
   protected Object invokeTransactionalMethod(GlobalTransaction tx, MethodCall m)
      throws Exception
   {
      Object retval = null;

      // 1. invoke method.
      try {
         retval = m.invoke(this, null);
      } catch (Throwable t) {
         retval=t;
      }

      // 2. If retval is timeout or locking exception:
      //    - revert all modifications done in the current TX
      //    - remove all temporary nodes created by the current TX
      //    - release all locks held by the current TX
      //    - return the exception
      if (retval != null && retval instanceof Exception)
      {
         rollback(tx);
         throw (Exception) retval;
      }
      else
      {
         // add the modification to the TX's modification list. this is used to later
         // (on TX commit) send all modifications done in this TX to all members
         tx_table.addModification(tx, m);
         return retval;
      }
   }

   protected Object invokeTransactionlessMethod(MethodCall m) throws Exception
   {
      List rsps;

      if (log.isDebugEnabled())
      {
         log.debug("invokeTransactionlessMethod(): invoking method " + m.getName() +
            "(" + m.getArgs() + ")" +
            ", members=" + members + ", synchronous=" + (cache_mode == REPL_SYNC) +
            ", exclude_self=" + true + ", timeout=" + sync_repl_timeout);
      }

      switch (cache_mode)
      {
         case REPL_ASYNC:

            // 1. Apply change locally
            try {
               m.invoke(this, null);
            }
            catch (Throwable t) {
               log.error("failed async replication", t);
            }

            // 2. Replicate change to all *other* members (exclude self !)
            if (use_repl_queue && repl_queue != null)
            {
               ReplicationQueue.Job j;
               j = new ReplicationQueue.Job(this, (Vector) members.clone(),
                  true, // exclude self ?
                  m.getMethod(), m.getArgs());
               repl_queue.add(j);
               try {
                  m.invoke(this, null);
               }
               catch (Throwable t) {
                  log.error("failed method invocation", t);
               }
            }
            else
               callRemoteMethods(members, m,
                  false, // synchronous ?
                  true, // exclude self ?
                  sync_repl_timeout);
            break;
         case REPL_SYNC:
            // REVISIT Needs to exclude itself and apply the local change manually.
            // This is needed such that transient field is modified properly in-VM.
            rsps = callRemoteMethods(members, m, true, true, sync_repl_timeout);
            if (log.isDebugEnabled())
               log.debug("invokeTransactionlessMethod(): responses=" + rsps);
            checkResponses(rsps);
            // 1. Apply change locally
            try {
               m.invoke(this, null);
            }
            catch (Throwable t) {
               log.error("failed invoking method", t);
            }
            break;
      }
      return null;
   }

   protected String printArgs(Object[] args)
   {
      StringBuffer sb = new StringBuffer();
      boolean first = true;
      if (args != null && args.length > 0)
      {
         for (int i = 0; i < args.length; i++)
         {
            if (!first)
               sb.append(", ");
            else
               first = false;
            if (args[i] instanceof String) sb.append("\"");
            sb.append(args[i]);
            if (args[i] instanceof String) sb.append("\"");
         }
      }
      return sb.toString();
   }

   /**
    * Finds a node given a fully qualified name. If intermediary nodes don't exist, they
    * will be created (if <tt>create_if_not_exists</tt> is true).
    * If <tt>lock_node</tt> is true, locks will be acquired for each node
    * (with <tt>tx</tt> as owner).<br>
    * Whenever nodes are created, and the global transaction is not null, the created
    * nodes have to be added to the transaction's {@link TransactionEntry#nodes}
    * field.<br>
    * When a lock is acquired on a node, a reference to the lock has to be added to the list
    * of locked nodes in the {@link TransactionEntry#locks} field.
    * <p>This operation will also applies different locking to the tree nodes, depending on
    * <tt>operation_type</tt>. If it is <tt>read</tt> type, all nodes will be acquired with
    * read lock. Otherwise, the operation is <tt>write</tt> type, all parent nodes will be acquired
    * with read lock while the destination node acquires write lock.</p>
    *
    * @param fqn                  Fully qualified name for the corresponding node.
    * @param create_if_not_exists Flag to indicate whether to create a new sub tree node if
    *                             not yet existed. Note that this applies only when lock_type is 1, i.e., read only.
    * @param lock_node            Boolean flag to decide whether to lock the node.
    * @param tx                   GlobaTransaction intance used to associated with this transaction.
    * @param operation_type       Integer to indicate whether this operation is for read or write
    *                             intention. Values are <tt>NODE_READ</tt> and <tt>NODE_WRITE</tt>.
    * @return
    * @throws LockingException
    * @throws TimeoutException
    */
   protected Node findNode(Fqn fqn, boolean create_if_not_exists, boolean lock_node,
      GlobalTransaction tx, int operation_type)
      throws LockingException, TimeoutException
   {
      Node n, child_node = null;
      Object child_name;
      Fqn tmp_fqn = new Fqn(), copy;
      int treeNodeSize = fqn.size();
      GlobalTransaction gtx = null;

      if (fqn == null) return null;
      if (treeNodeSize == 0)
         return root;
      n = root;
      // Check if it exists first.
      if (create_if_not_exists == false)
      {
         for (int i = 0; i < treeNodeSize; i++)
         {
            child_node = n.getChild(fqn.get(i));
            if (child_node == null) return null;
            n = child_node;
         }
      }

      n = root;
      for (int i = 0; i < treeNodeSize; i++)
      {
         child_name = fqn.get(i);
         tmp_fqn.add(child_name);
         child_node = n.getChild(child_name);
         if (child_node == null)
         {
            copy = (Fqn) tmp_fqn.clone();
            child_node = n.createChild(child_name, copy, n);
            if (tx != null)
            {
               // add the node name to the list maintained for the current tx
               // (needed for abort/rollback of transaction)
               tx_table.addNode(tx, (Fqn) tmp_fqn.clone());
            }
            notifyNodeAdded(copy);
         }

         if (lock_node)
         {
            if (tx == null)
            {
               gtx = getGlobalTransactionFromThread();
            }
            else
            {
               gtx = tx;
            }

            try
            {
               // Write lock only for destination node in write request.
               if (operation_type == NODE_WRITE && i == (treeNodeSize - 1))
               {
                  child_node.acquire(gtx, lock_acquisition_timeout, Node.LOCK_TYPE_WRITE);
               }
               else
               {
                  child_node.acquire(gtx, lock_acquisition_timeout, Node.LOCK_TYPE_READ);
               }
            }
            catch (LockingException cex)
            {
               releaseLocks(gtx, fqn); // This is intentional. If tx is null, releaseLocks will understand it.
               throw cex;
            }
            catch (TimeoutException cex)
            {
               releaseLocks(gtx, fqn);
               throw cex;
            }

            if (tx != null)
            {
               // add the lock to the list of locks maintained for this transaction
               // (needed for release of locks on commit or rollback)
               tx_table.addLock(tx, child_node.getLock());
            }
         }
         n = child_node;
      }
      return child_node;
   }

   /**
    * This method finds a node given a fully qualified name. It <em>never</em> creates
    * intermediary nodes.
    *
    * @param fqn
    * @param lock_type Lock type. Read lock --1, write lock --2.
    * @return
    */
   protected Node findNode(String fqn, int lock_type) throws LockingException, TimeoutException
   {
      return findNode(Fqn.fromString(fqn), false, locking, getCurrentTransaction(), lock_type);
   }

   /**
    * This method finds a node given a fully qualified name. It <em>never</em> creates
    * intermediary nodes.
    *
    * @param fqn
    * @param lock_type Lock type. Read lock --1, write lock --2.
    * @return
    */
   protected Node findNode(Fqn fqn, int lock_type) throws LockingException, TimeoutException
   {
      return findNode(fqn, false, locking, getCurrentTransaction(), lock_type);
   }

   /**
    * Obtain a GlobalTransaction instance based on current thread id.
    * Note that once a gtx is mapped the a thread instance, it will never
    * be removed from the map. In other words, the same thread will use
    * the same gtx instance during the cache life cycle.
    *
    * @return GlobalTransaction instance
    */
   protected GlobalTransaction getGlobalTransactionFromThread()
   {
      Object currentThread = Thread.currentThread();
      GlobalTransaction gtx = (GlobalTransaction) thread2GtxMap_.get(currentThread);
      if (gtx == null)
      {
         Address ipAddress = (Address) getLocalAddress();
         gtx = GlobalTransaction.create(ipAddress);
         thread2GtxMap_.put(currentThread, gtx);
      }

      return gtx;
   }

   /**
    * Releases the locks acquired outside of a transaction
    *
    * @param fqn
    */
   protected void releaseLocks(GlobalTransaction tx, Fqn fqn)
   {
      Node n, child_node = null;
      Object child_name;

      if (fqn == null) return;
      n = root;

      // TODO tx is kind of confusing here. Need to consolidate it later.
      if (tx == null)
      { // Use current thread to identify the GlobalTransaction
         tx = getGlobalTransactionFromThread();
      }

      java.util.Stack stack = new java.util.Stack();
      for (int i = 0; i < fqn.size(); i++)
      {
         child_name = fqn.get(i);
         child_node = n.getChild(child_name);
         if (child_node == null)
         {
            return;
         }
         else
         {
            stack.push(child_node);
         }
         n = child_node;
      }

      while (!stack.empty())
      {
         Node node = (Node) stack.pop();
         node.release(tx);
      }
   }

   /**
    * Returns the data hashmap for a given node. This method can only be used by callers that are inside
    * the same package. The reason is that callers must not modify the return value, as these modifications
    * would not be replicated, thus rendering the replicas inconsistent.
    *
    * @param fqn The fully qualified name of the node
    * @return Map A cloned data hashmap for the given node
    */
   protected Map getData(String fqn) throws LockingException, TimeoutException
   {
      Node n = findNode(fqn, NODE_READ);

      if (n == null) return null;
      return n.getData();
   }


   protected void notifyNodeAdded(Fqn fqn)
   {
      for (int i = 0; i < listeners.size(); i++)
         ((TreeCacheListener) listeners.elementAt(i)).nodeAdded(fqn);
   }

   protected void notifyNodeRemoved(Fqn fqn)
   {
      for (int i = 0; i < listeners.size(); i++)
         ((TreeCacheListener) listeners.elementAt(i)).nodeRemoved(fqn);
   }

   protected void notifyNodeModified(Fqn fqn)
   {
      for (int i = 0; i < listeners.size(); i++)
         ((TreeCacheListener) listeners.elementAt(i)).nodeModified(fqn);
   }

   protected void notifyNodeVisisted(Fqn fqn)
   {
      for (int i = 0; i < listeners.size(); i++)
         ((TreeCacheListener) listeners.elementAt(i)).nodeVisited(fqn);
   }

   protected void notifyCacheStarted()
   {
      for (int i = 0; i < listeners.size(); i++)
         ((TreeCacheListener) listeners.elementAt(i)).cacheStarted(this);
   }

   protected void notifyCacheStopped()
   {
      for (int i = 0; i < listeners.size(); i++)
         ((TreeCacheListener) listeners.elementAt(i)).cacheStopped(this);
   }

   protected void notifyViewChange(View v)
   {
      for (int i = 0; i < listeners.size(); i++)
         ((TreeCacheListener) listeners.elementAt(i)).viewChange(v);
   }

   /**
    * Generates NodeAdded notifications for all nodes of the tree. This is called whenever the tree is
    * initially retrieved (state transfer)
    */
   protected void notifyAllNodesCreated(Node curr)
   {
      Node n;
      Map children;

      if (curr == null) return;
      notifyNodeAdded(curr.fqn);
      if ((children = curr.getChildren()) != null)
      {
         for (Iterator it = children.values().iterator(); it.hasNext();)
         {
            n = (Node) it.next();
            notifyAllNodesCreated(n);
         }
      }
   }

   protected String getDefaultProperties()
   {
      return "UDP(mcast_addr=224.0.0.36;mcast_port=55566;ip_ttl=32;" +
         "mcast_send_buf_size=150000;mcast_recv_buf_size=80000):" +
         "PING(timeout=2000;num_initial_members=3):" +
         "MERGE2(min_interval=5000;max_interval=10000):" +
         "FD_SOCK:" +
         "VERIFY_SUSPECT(timeout=1500):" +
         "pbcast.STABLE(desired_avg_gossip=20000):" +
         "pbcast.NAKACK(gc_lag=50;retransmit_timeout=600,1200,2400,4800):" +
         "UNICAST(timeout=5000):" +
         "FRAG(frag_size=16000;down_thread=false;up_thread=false):" +
         "pbcast.GMS(join_timeout=5000;join_retry_timeout=2000;" +
         "shun=false;print_local_addr=true):" +
         "pbcast.STATE_TRANSFER";
   }

   /**
    * Checks whether any of the responses are exceptions. If yes, re-throws
    * them (as exceptions or runtime exceptions).
    *
    * @param rsps
    * @throws Exception
    */
   protected void checkResponses(List rsps) throws Exception
   {
      Object rsp;
      if (rsps != null)
      {
         for (Iterator it = rsps.iterator(); it.hasNext();)
         {
            rsp = it.next();
            if (rsp != null)
            {
               if (rsp instanceof RuntimeException)
                  throw (RuntimeException) rsp;
               if (rsp instanceof Exception)
                  throw (Exception) rsp;
            }
         }
      }
   }


   protected int string2Mode(String mode)
   {
      if (mode == null) return -1;
      String m = mode.toLowerCase().trim();
      if (m.equals("local"))
         return LOCAL;
      else if (m.equals("repl_async") || m.equals("repl-async"))
         return REPL_ASYNC;
      else if (m.equals("repl_sync") || m.equals("repl-sync"))
         return REPL_SYNC;
      else
         return -1;
   }


   int numberOfModifications(GlobalTransaction gtx)
   {
      TransactionEntry entry;
      if (gtx == null)
         return -1;
      entry = tx_table.get(gtx);
      List modifications = entry != null ? entry.getModifications() : null;
      return modifications != null ? modifications.size() : -1;
   }

   /**
    * This class handles the replication of all modifications performed under a transaction
    * when the transaction completes
    */
   class SynchronizationHandler implements Synchronization
   {
      GlobalTransaction gtx = null;
      TreeCache tc = null;

      SynchronizationHandler(GlobalTransaction gtx, TreeCache tc)
      {
         this.gtx = gtx;
         this.tc = tc;
      }

      /**
       * This method is invoked before the start of the commit or rollback
       * process. The method invocation is done in the context of the
       * transaction that is about to be committed or rolled back. We need to check
       * whether the TX is marked as rolled back (then we need to roll back as well)
       * or whether it is okay. If okay we start our own 2-phase commit protocol.
       * Any exception thrown will be setting the status of the transaction to
       * marked as rollback.
       */
      public void beforeCompletion()
      {
         Transaction tx = getLocalTransaction();

         // this is a no-op when we replicate asynchronously or when mode is local
         if (cache_mode == LOCAL || cache_mode == REPL_ASYNC)
            return;

         try
         {
            int status = tx.getStatus();
            if (log.isDebugEnabled())
               log.debug("beforeCompletion() called (status=" + status + ")");

            // we don't replicate if there are no modifications in the current transaction
            if (numberOfModifications(gtx) == 0)
            {
               if (log.isDebugEnabled())
                  log.debug("beforeCompletion(): no modifications in this transaction");
               return;
            }

            switch (status)
            {
               case Status.STATUS_ACTIVE: // added Feb 2 2004 (bela)
               case Status.STATUS_COMMITTING:
                  //case Status.STATUS_PREPARING: preparing is an illegal state in a synchronization
                  // TX okay, run PREPARE phase. Any exception will cause this TX to roll back
                  try
                  {
                     runPreparePhase(gtx,
                        (Address) getLocalAddress(),
                        false, // synchronous
                        false);  // don't commit or rollback - wait for call
                  }
                  catch (Exception t)
                  {
                     log.warn("beforeCompletion(): runPreparePhase failed. Action is now rollback." +
                        " Exception " + t);
                     // rollback(gtx); // changed by Bela Feb 5 2004: we can do this in the afterCompletion()

                     // change Bela Nov 5 2003: we need to rethrow the exception so the Transaction
                     // knows it has to set status of TX to rolled back
                     throw t;
                  }
                  break;
               case Status.STATUS_MARKED_ROLLBACK:
               case Status.STATUS_ROLLING_BACK:
                  // rollback(gtx); // changed by Bela Feb 5 2004: we can do this in the afterCompletion()
                  break;
               default:
                  log.error("beforeCompletion(). Illegal tx status: " + status);
                  throw new IllegalStateException("Illegal status: " + status);
            }
         }
         catch (Throwable t)
         {
            throw new NestedRuntimeException("", t);
         }
      }


      /**
       * This method is invoked after the transaction has committed or
       * rolled back.
       * <p/>
       * Um, how can this work? the only possible input status
       * values are Committed and RolledBack, which both result in
       * throwing an illegal state exception!
       *
       * @param status The status of the completed transaction.
       */
      public void afterCompletion(int status)
      {
         if (log.isDebugEnabled())
            log.debug("afterCompletion() called (status=" + status + ")");

         switch (status)
         {
// BW
//                case Status.STATUS_ACTIVE:
//                case Status.STATUS_PREPARED:
//                case Status.STATUS_COMMITTING:
            case Status.STATUS_COMMITTED:

               switch (cache_mode)
               {
                  case LOCAL:
                     commit(gtx);
                     break;
                  case REPL_ASYNC:
                     if (numberOfModifications(gtx) > 0)
                     {
                        try
                        {
                           if (use_repl_queue && repl_queue != null)
                           {
                              TransactionEntry entry = tx_table.get(gtx);
                              if (entry == null)
                                 throw new Exception("runPreparePhase(): found no entry for transaction " + gtx);
                              List modifications = entry.getModifications();
                              ReplicationQueue.Job j;
                              j = new ReplicationQueue.Job(tc,
                                 (Vector) members.clone(),
                                 true, // exclude self
                                 prepare_method,
                                 new Object[]{gtx, modifications,
                                              (Address) getLocalAddress(),
                                              new Boolean(true)});
                              repl_queue.add(j);
                           }
                           else
                           {
                              // asynchronous PREPARE multicast, excluding self
                              runPreparePhase(gtx,
                                 (Address) getLocalAddress(),
                                 true, // async
                                 true);  // do commit or rollback after call
                           }
                        }
                        catch (Throwable t)
                        {
                           log.warn("afterCompletion()", t);
                        }
                     }
                     commit(gtx);
                     break;
                  case REPL_SYNC:
                     runCommitPhase(gtx); // commit locally, then send async commit() message to other mbrs
                     break;
               }
               break;

            case Status.STATUS_MARKED_ROLLBACK: // this one is probably not needed
            case Status.STATUS_ROLLEDBACK:
               if (log.isDebugEnabled())
                  log.debug("afterCompletion(): rolling back cache");
               switch (cache_mode)
               {
                  case LOCAL:
                  case REPL_ASYNC:
                     rollback(gtx); // roll back locally
                     break;
                  case REPL_SYNC:
                     runRollbackPhase(gtx);
                     break;
               }
               break;
            default:
               switch (cache_mode)
               {
                  case LOCAL:
                  case REPL_ASYNC:
                     rollback(gtx); // roll back locally
                     break;
                  case REPL_SYNC:
                     runRollbackPhase(gtx);
                     break;
               }
               throw new IllegalStateException("Illegal status: " + status);
         }
      }
   }

   static void help()
   {
      System.out.println("TreeCache [-help] [-cluster_props <properties>] [-mode <mode>]");
      System.out.println("(mode: local (default), repl_async, repl_sync");
   }

   public static class MyListener implements TreeCacheListener
   {

      public void nodeAdded(Fqn fqn)
      {
         System.out.println("** node added: " + fqn);
      }

      public void nodeRemoved(Fqn fqn)
      {
         System.out.println("** node removed: " + fqn);
      }

      public void nodeModified(Fqn fqn)
      {
         System.out.println("** node modified: " + fqn);
      }

      public void nodeVisited(Fqn fqn)
      {
      }

      public void cacheStarted(TreeCache cache)
      {
         //To change body of implemented methods use File | Settings | File Templates.
      }

      public void cacheStopped(TreeCache cache)
      {
         //To change body of implemented methods use File | Settings | File Templates.
      }

      public void viewChange(View new_view)
      {
         System.out.println("** view change: " + new_view);
      }
   }


}


