/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 * Created on March 25 2003
 */
package org.jboss.cache.eviction;

import EDU.oswego.cs.dl.util.concurrent.BoundedLinkedQueue;
import org.jboss.cache.Fqn;

/**
 * A region is a collection of tree cache nodes that share the same eviction
 * policy configurations, e.g., maxNodes, etc. The region is specified via
 * Fqn.
 *
 * @author Ben Wang 2-2004
 */
public class Region
{
   // Signal the operation types.
   public static final Integer ADD_EVENT = new Integer(0);
   public static final Integer REMOVE_EVENT = new Integer(1);
   public static final Integer VISIT_EVENT = new Integer(2);

   // Policy-specific config
   private int maxNodes_ = 0; // 0 is no limit
   private long timeToIdleSeconds_ = 0; // 0 is forever
   // Region name
   private String fqn_;
   // Policy-spefic algorithm, e.g., LRU
   private EvictionAlgorithm algorithm_;
   // Parent policy provider
   private EvictionPolicy policy_;
   // These queues can do put and take concurrently.
   protected BoundedLinkedQueue nodeQueue_;
   protected BoundedLinkedQueue eventQueue_;

   /**
    * Default to package namespace on purpose so no one outside the package can instantiate it,
    */
   Region()
   {
      createQueues();
   }

   void createQueues()
   {
      nodeQueue_ = new BoundedLinkedQueue(RegionManager.CAPACITY);
      eventQueue_ = new BoundedLinkedQueue(RegionManager.CAPACITY);
   }

   Region(String fqn, EvictionPolicy policy, EvictionAlgorithm algorithm)
   {
      fqn_ = fqn;
      policy_ = policy;
      algorithm_ = algorithm;

      createQueues();
   }

   public int getMaxNodes()
   {
      return maxNodes_;
   }

   public void setMaxNodes(int maxNodes)
   {
      maxNodes_ = maxNodes;
   }

   public long getTimeToIdleSeconds()
   {
      return timeToIdleSeconds_;
   }

   public void setTimeToIdleSeconds(long secs)
   {
      timeToIdleSeconds_ = secs;
   }

   public EvictionAlgorithm getEvictionAlgorithm()
   {
      return algorithm_;
   }

   public EvictionPolicy getEvictionPolicy()
   {
      return policy_;
   }

   public String getFqn()
   {
      return fqn_;
   }


   public void setAddedNode(Fqn fqn)
   {
      try
      {
         nodeQueue_.put(fqn);
         eventQueue_.put(ADD_EVENT);
      }
      catch (InterruptedException e)
      {
         e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
      }
   }

   public void setRemovedNode(Fqn fqn)
   {
      try
      {
         nodeQueue_.put(fqn);
         eventQueue_.put(REMOVE_EVENT);
      }
      catch (InterruptedException e)
      {
         e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
      }
   }

   public void setVisitedNode(Fqn fqn)
   {
      try
      {
         nodeQueue_.put(fqn);
         eventQueue_.put(VISIT_EVENT);
      }
      catch (InterruptedException e)
      {
         e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
      }
   }

   /**
    * Take the last node from node queue. It will also
    * remove it from the queue.
    * @return
    */
   public Fqn takeLastNode()
   {
      try
      {
         return (Fqn) nodeQueue_.poll(0);
      }
      catch (InterruptedException e)
      {
         e.printStackTrace();
      }
      return null;
   }

   /**
    * Take the last node from event queue. It will also
    * remove it from the queue.
    * @return
    */
   public Integer takeLastEvent()
   {
      try
      {
         return (Integer) eventQueue_.poll(0);
      }
      catch (InterruptedException e)
      {
         e.printStackTrace();
      }
      return null;
   }

   public int nodeQueueSize()
   {
      return nodeQueue_.size();
   }

   public int eventQueueSize()
   {
      return eventQueue_.size();
   }

   public String toString()
   {
      StringBuffer buf = new StringBuffer();
      buf.append("Regions--- fqn: ").append(getFqn()).append(" maxNodes " + getMaxNodes());
      buf.append(" TimeToIdleSeconds " + getTimeToIdleSeconds());
      return buf.toString();
   }
}
