/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.lock;

import java.util.ArrayList;
import java.util.Collection;

/**
 * Provide lock ownership mapping.
 *
 * @author Ben
 */
public class LockMap
{
   public static final int OWNER_ANY = 0;
   public static final int OWNER_READ = 1;
   public static final int OWNER_WRITE = 2;
   private Object writeOwner_;
   private Collection readOwnerList_;

   public LockMap()
   {
      writeOwner_ = null;
      readOwnerList_ = new ArrayList();
   }

   /**
    * Check whether this owner has reader or writer ownership.
    *
    * @param owner
    * @param ownership Either <code>OWNER_ANY</code>, <code>OWNER_READ</code>,
    *                  or <code>OWNER_WRITE</code>.
    * @return
    */
   public synchronized boolean isOwner(Object owner, int ownership)
   {
      checkOwnerNull(owner);
      boolean hasOwner = false;
      switch (ownership)
      {
         case OWNER_ANY:
            if ((writeOwner_ != null && writeOwner_.equals(owner) ||
               readOwnerList_.size() != 0 && readOwnerList_.contains(owner)))
               hasOwner = true;
            break;
         case OWNER_READ:
            if (readOwnerList_.size() != 0 && readOwnerList_.contains(owner))
               hasOwner = true;
            break;
         case OWNER_WRITE:
            if (writeOwner_ != null && writeOwner_.equals(owner)) hasOwner = true;
            break;
         default:
            break;
      }
      return hasOwner;
   }

   private void checkOwnerNull(Object owner)
   {
      if (owner == null)
         throw new IllegalArgumentException("Owner object is null");
   }

   /**
    * Adding a reader owner.
    *
    * @param owner
    */
   public synchronized void addReader(Object owner)
   {
      checkOwnerNull(owner);
      readOwnerList_.add(owner);
   }

   /**
    * Adding a writer owner.
    *
    * @param owner
    */
   public synchronized void addWriter(Object owner)
   {
      checkOwnerNull(owner);
      if (writeOwner_ != null)
         throw new IllegalStateException("addWriter(): owner already existed");

      writeOwner_ = owner;
   }

   /**
    * Upgrading current reader ownership to writer one.
    *
    * @param owner
    * @return True if successful.
    */
   public synchronized boolean upgrade(Object owner) throws OwnerNotExistedException
   {
      checkOwnerNull(owner);
      if (!readOwnerList_.contains(owner))
         throw new OwnerNotExistedException("Can't upgrade lock. Read lock owner not existed");

      readOwnerList_.remove(owner);
      writeOwner_ = owner;
      return true;
   }

   /**
    * @return Collection of reader owner objects. Size 0 if not found.
    */
   public synchronized Collection readerOwners()
   {
      return readOwnerList_;
   }

   /**
    * @return Writer owner object. Null if none.
    */
   public synchronized Object writerOwner()
   {
      return writeOwner_;
   }

   /**
    * Remove reader ownership.
    *
    * @param owner
    * @return The owner object. Null if not found.
    */
   public synchronized boolean removeReader(Object owner)
   {
      checkOwnerNull(owner);
      return readOwnerList_.remove(owner);
   }

   /**
    * Remove writer ownership.
    *
    * @param owner
    * @return The owner object. Null if not found.
    */
   public synchronized Object removeWriter(Object owner)
   {
      checkOwnerNull(owner);
      Object tmp = writeOwner_;
      writeOwner_ = null;
      return tmp;
   }

   /**
    * Remove all ownership.
    */
   public synchronized void removeAll()
   {
      writeOwner_ = null;
      readOwnerList_.clear();
   }

   /**
    * Debugging information.
    *
    * @return
    */
   public String printInfo()
   {
      StringBuffer buf = new StringBuffer();
      buf.append("Read lock owners: ").append(readOwnerList_).append("\n");
      buf.append("Write lock owner: ").append(writeOwner_).append("\n");
      return buf.toString();
   }
}
