/*
 *
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.test.cache.perf;


import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import org.jboss.cache.TreeCache;
import org.jboss.cache.lock.LockStrategyFactory;
import org.jboss.cache.lock.IsolationLevel;
import org.jboss.cache.transaction.DummyTransactionManager;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.transaction.UserTransaction;
import java.text.DecimalFormat;
import java.text.FieldPosition;
import java.util.ArrayList;
import java.util.Properties;

/**
 * Replicated synchronous mode performance test for transactional TreeCache.
 * @author<a href="mailto:bwang@jboss.org">Ben Wang</a> May 20 2003
 * @version $Revision: 1.6.2.1 $
 */
public class ReplicatedSyncPerfTestCase extends TestCase {
   TreeCache           cache1_, cache2_, cache3_;
   int                 cachingMode_ = TreeCache.REPL_SYNC;
   final String              groupName_ = "TreeCacheTestGroup";
   final static Properties p_;
//   final static Logger log_ = Logger.getLogger(ReplicatedSyncPerfTestCase.class);
   String        oldFactory_ = null;
   final String  FACTORY = "org.jboss.cache.transaction.DummyContextFactory";
    String props_ =
        "UDP(ip_mcast=true;ip_ttl=64;loopback=false;mcast_addr=228.1.2.3;" +
            "mcast_port=45566;mcast_recv_buf_size=80000;mcast_send_buf_size=150000;" +
            "ucast_recv_buf_size=80000;ucast_send_buf_size=150000):" +
        "PING(down_thread=true;num_initial_members=3;timeout=2000;up_thread=true):" +
        "MERGE2(max_interval=20000;min_interval=10000):" +
        "FD(down_thread=true;shun=true;up_thread=true):" +
        "VERIFY_SUSPECT(down_thread=true;timeout=1500;up_thread=true):" +
        "pbcast.NAKACK(down_thread=true;gc_lag=50;retransmit_timeout=600,1200,2400,4800;" +
            "up_thread=true):" +
        "pbcast.STABLE(desired_avg_gossip=20000;down_thread=true;up_thread=true):" +
        "UNICAST(down_thread=true;min_threshold=10;timeout=600,1200,2400;window_size=100):" +
        "FRAG(down_thread=true;frag_size=8192;up_thread=true):" +
        "pbcast.GMS(join_retry_timeout=2000;join_timeout=5000;print_local_addr=true;shun=true):" +
        "pbcast.STATE_TRANSFER(down_thread=true;up_thread=true)";

   ArrayList nodeList_;
   // (4, 4) combination will generate 340 nodes.
   static final int depth_ = 4;
   static final int children_ = 4;
   DummyTransactionManager tm_;

   static {
      p_=new Properties();
      p_.put(Context.INITIAL_CONTEXT_FACTORY, "org.jboss.cache.transaction.DummyContextFactory");
   }

   public ReplicatedSyncPerfTestCase(String name) {
      super(name);
   }

   public void setUp() throws Exception {
      super.setUp();

      oldFactory_ = System.getProperty(Context.INITIAL_CONTEXT_FACTORY);
      System.setProperty(Context.INITIAL_CONTEXT_FACTORY, FACTORY);

      DummyTransactionManager.getInstance();
      nodeList_ = nodeGen(depth_, children_);
      tm_ = new DummyTransactionManager();

      log("ReplicatedSyncPerfTestCase: cacheMode=REPL_SYNC");
   }

   public void tearDown() throws Exception {
      super.tearDown();

      DummyTransactionManager.destroy();

      if(oldFactory_ != null) {
         System.setProperty(Context.INITIAL_CONTEXT_FACTORY, oldFactory_);
         oldFactory_ = null;
      }

   }

   TreeCache createCache() throws Exception {
      TreeCache cache = new TreeCache(groupName_, props_, 10000);
      cache.setCacheMode(cachingMode_);
      cache.setIsolationLevel(IsolationLevel.REPEATABLE_READ);
      cache.setTransactionManagerLookupClass("org.jboss.cache.JBossTransactionManagerLookup");
      return cache;
   }

   void destroyCache(TreeCache cache) throws Exception {
      cache.stop();
      cache = null;
   }

   public void testOneCacheTx() throws Exception {
      log("=== 1 cache with transaction (no concurrent access) ===");
      cache1_ = createCache();
      cache1_.setTransactionManagerLookupClass("org.jboss.cache.JBossTransactionManagerLookup");
      cache1_.start();

      // Formating
      DecimalFormat form = new DecimalFormat("#.00");
      FieldPosition fieldPos = new FieldPosition(0);
      StringBuffer dumbStr = new StringBuffer();
      boolean hasTx = true;
      boolean oneTxOnly = false;

      // Step 1. Add entries to the cache
      long time1 = System.currentTimeMillis();
      int nOps = _add(cache1_, hasTx, oneTxOnly);
      long time2 = System.currentTimeMillis();
      double d = (double)(time2-time1)/nOps;
      log("Time elapsed for _add is " + (time2-time1) + " with " +nOps
         +" operations. Average per ops is: " + form.format(d, dumbStr, fieldPos) +
         " msec." );
      dumbStr = new StringBuffer();

      // Step 2. Query the cache
      time1 = System.currentTimeMillis();
      nOps = _get(cache1_, hasTx, oneTxOnly);
      time2 = System.currentTimeMillis();
      d = (double)(time2-time1)/nOps;
      log("Time elapsed for _get is " + (time2-time1) + " with " +nOps
         +" operations. Average per ops is: " + form.format(d, dumbStr, fieldPos) +
         " msec." );
      dumbStr = new StringBuffer();

      // Step 3. Remove entries from the cache
      time1 = System.currentTimeMillis();
      nOps = _remove(cache1_, hasTx, oneTxOnly);
      time2 = System.currentTimeMillis();
      d = (double)(time2-time1)/nOps;
      log("Time elapsed for _remove is " + (time2-time1) + " with " +nOps
         +" operations. Average per ops is: " + form.format(d, dumbStr, fieldPos) +
         " msec." );

      destroyCache(cache1_);
   }

   public void test2CachesTx() throws Exception {
      log("=== 2 caches with transaction (no concurrent access) ===");
      cache1_ = createCache();
      cache2_ = createCache();
      cache1_.setTransactionManagerLookupClass("org.jboss.cache.JBossTransactionManagerLookup");
      cache2_.setTransactionManagerLookupClass("org.jboss.cache.JBossTransactionManagerLookup");
      cache1_.start();
      cache2_.start();

      // Formating
      DecimalFormat form = new DecimalFormat("#.00");
      FieldPosition fieldPos = new FieldPosition(0);
      StringBuffer dumbStr = new StringBuffer();
      boolean hasTx = true;
      boolean oneTxOnly = false;

      // Step 1. Add entries to the cache
      long time1 = System.currentTimeMillis();
      int nOps = _add(cache1_, hasTx, oneTxOnly);
      long time2 = System.currentTimeMillis();
      double d = (double)(time2-time1)/nOps;
      log("Time elapsed for _add is " + (time2-time1) + " with " +nOps
         +" operations. Average per ops is: " + form.format(d, dumbStr, fieldPos) +
         " msec." );
      dumbStr = new StringBuffer();

      // Step 2. Query the cache
      time1 = System.currentTimeMillis();
      nOps = _get(cache1_, hasTx, oneTxOnly);
      time2 = System.currentTimeMillis();
      d = (double)(time2-time1)/nOps;
      log("Time elapsed for _get is " + (time2-time1) + " with " +nOps
         +" operations. Average per ops is: " + form.format(d, dumbStr, fieldPos) +
         " msec." );
      dumbStr = new StringBuffer();

      // Step 3. Remove entries from the cache
      time1 = System.currentTimeMillis();
      nOps = _remove(cache2_, hasTx, oneTxOnly);  // Note we remove nodes from cache2.
      time2 = System.currentTimeMillis();
      d = (double)(time2-time1)/nOps;
      log("Time elapsed for _remove is " + (time2-time1) + " with " +nOps
         +" operations. Average per ops is: " + form.format(d, dumbStr, fieldPos) +
         " msec." );

      destroyCache(cache1_);
      destroyCache(cache2_);
   }

   public void test2CachesOneTxOnly() throws Exception {
      log("=== 2 caches with single transaction only (no concurrent access) ===");
      cache1_ = createCache();
      cache2_ = createCache();
      cache1_.setTransactionManagerLookupClass("org.jboss.cache.JBossTransactionManagerLookup");
      cache2_.setTransactionManagerLookupClass("org.jboss.cache.JBossTransactionManagerLookup");
      cache1_.start();
      cache2_.start();

      // Formating
      DecimalFormat form = new DecimalFormat("#.00");
      FieldPosition fieldPos = new FieldPosition(0);
      StringBuffer dumbStr = new StringBuffer();
      boolean hasTx = true;
      boolean oneTxOnly = true;

      // Step 1. Add entries to the cache
      long time1 = System.currentTimeMillis();
      int nOps = _add(cache1_, hasTx, oneTxOnly);
      long time2 = System.currentTimeMillis();
      double d = (double)(time2-time1)/nOps;
      log("Time elapsed for _add is " + (time2-time1) + " with " +nOps
         +" operations. Average per ops is: " + form.format(d, dumbStr, fieldPos) +
         " msec." );
      dumbStr = new StringBuffer();

      // Step 2. Query the cache
      time1 = System.currentTimeMillis();
      nOps = _get(cache1_, hasTx, oneTxOnly);
      time2 = System.currentTimeMillis();
      d = (double)(time2-time1)/nOps;
      log("Time elapsed for _get is " + (time2-time1) + " with " +nOps
         +" operations. Average per ops is: " + form.format(d, dumbStr, fieldPos) +
         " msec." );
      dumbStr = new StringBuffer();

      // Step 3. Remove entries from the cache
      time1 = System.currentTimeMillis();
      nOps = _remove(cache2_, hasTx, oneTxOnly);  // Note we remove nodes from cache2.
      time2 = System.currentTimeMillis();
      d = (double)(time2-time1)/nOps;
      log("Time elapsed for _remove is " + (time2-time1) + " with " +nOps
         +" operations. Average per ops is: " + form.format(d, dumbStr, fieldPos) +
         " msec." );

      destroyCache(cache1_);
      destroyCache(cache2_);
   }

   public void test3CachesTx() throws Exception {
      log("=== 3 caches with transaction (no concurrent access) ===");
      cache1_ = createCache();
      cache2_ = createCache();
      cache3_ = createCache();
      cache1_.setTransactionManagerLookupClass("org.jboss.cache.JBossTransactionManagerLookup");
      cache2_.setTransactionManagerLookupClass("org.jboss.cache.JBossTransactionManagerLookup");
      cache2_.setTransactionManagerLookupClass("org.jboss.cache.JBossTransactionManagerLookup");
      cache3_.setTransactionManagerLookupClass("org.jboss.cache.JBossTransactionManagerLookup");
      cache1_.start();
      cache2_.start();
      cache3_.start();

      // Formating
      DecimalFormat form = new DecimalFormat("#.00");
      FieldPosition fieldPos = new FieldPosition(0);
      StringBuffer dumbStr = new StringBuffer();
      boolean hasTx = true;
      boolean oneTxOnly = false;

      // Step 1. Add entries to the cache
      long time1 = System.currentTimeMillis();
      int nOps = _add(cache1_, hasTx, oneTxOnly);
      long time2 = System.currentTimeMillis();
      double d = (double)(time2-time1)/nOps;
      log("Time elapsed for _add is " + (time2-time1) + " with " +nOps
         +" operations. Average per ops is: " + form.format(d, dumbStr, fieldPos) +
         " msec." );
      dumbStr = new StringBuffer();

      // Step 2. Query the cache
      time1 = System.currentTimeMillis();
      nOps = _get(cache2_, hasTx, oneTxOnly);   // Note query is from cache2
      time2 = System.currentTimeMillis();
      d = (double)(time2-time1)/nOps;
      log("Time elapsed for _get is " + (time2-time1) + " with " +nOps
         +" operations. Average per ops is: " + form.format(d, dumbStr, fieldPos) +
         " msec." );
      dumbStr = new StringBuffer();

      // Step 3. Remove entries from the cache
      time1 = System.currentTimeMillis();
      nOps = _remove(cache3_, hasTx, oneTxOnly);   // Note remove is from cache3
      time2 = System.currentTimeMillis();
      d = (double)(time2-time1)/nOps;
      log("Time elapsed for _remove is " + (time2-time1) + " with " +nOps
         +" operations. Average per ops is: " + form.format(d, dumbStr, fieldPos) +
         " msec." );

      destroyCache(cache1_);
      destroyCache(cache2_);
      destroyCache(cache3_);
   }

   private int _add(TreeCache cache, boolean hasTx, boolean oneTxOnly) throws Exception {
      UserTransaction tx = null;
      if(hasTx) {
         tx=(UserTransaction)new InitialContext(p_).lookup("UserTransaction");
      }

      if( hasTx && oneTxOnly ) {
          tx.begin();
      }

      for(int i=0; i < nodeList_.size(); i++) {
         String key = Integer.toString(i);
         String value = Integer.toString(i);
         if(hasTx && !oneTxOnly) {
            tx.begin();
            cache.put((String)nodeList_.get(i), key, value);
            tx.commit();
         } else {
            cache.put((String)nodeList_.get(i), key, value);
         }
      }

      if( hasTx && oneTxOnly ) {
          tx.commit();
      }

       return nodeList_.size();
   }

   private int _get(TreeCache cache, boolean hasTx, boolean oneTxOnly) throws Exception {
      UserTransaction tx = null;
      if(hasTx) {
         tx=(UserTransaction)new InitialContext(p_).lookup("UserTransaction");
      }

      if( hasTx && oneTxOnly ) {
          tx.begin();
      }

      for(int i=0; i < nodeList_.size(); i++) {
         String key = Integer.toString(i);
         if(hasTx && !oneTxOnly) {
            tx.begin();
            cache.get((String)nodeList_.get(i), key);
            tx.commit();
         } else {
            cache.get((String)nodeList_.get(i), key);
         }
      }

      if( hasTx && oneTxOnly ) {
          tx.commit();
      }

      return nodeList_.size();
   }

   private int _remove(TreeCache cache, boolean hasTx, boolean oneTxOnly) throws Exception {
      UserTransaction tx = null;
      if(hasTx) {
         tx=(UserTransaction)new InitialContext(p_).lookup("UserTransaction");
      }

      if( hasTx && oneTxOnly ) {
          tx.begin();
      }

      for(int i=0; i < nodeList_.size(); i++) {
         String key = Integer.toString(i);
         if(hasTx && !oneTxOnly) {
            tx.begin();
            cache.remove((String)nodeList_.get(i), key);
            tx.commit();
         } else {
            cache.remove((String)nodeList_.get(i), key);
         }
      }

      if( hasTx && oneTxOnly ) {
          tx.commit();
      }

      return nodeList_.size();
   }

   /**
    * Generate the tree nodes quasi-exponentially. I.e., depth is the level
    * of the hierarchy and children is the number of children under each node.
    *
    */
   private ArrayList nodeGen(int depth, int children) {
      ArrayList strList = new ArrayList();
      ArrayList oldList = new ArrayList();
      ArrayList newList = new ArrayList();

      oldList.add("/");
      newList.add("/");
      strList.add("/");

      while(depth > 0) {
         // Trying to produce node name at this depth.
         newList = new ArrayList();
         for(int i=0; i < oldList.size(); i++) {
            for(int j=0; j < children; j++) {
               String tmp = (String)oldList.get(i);
               tmp += Integer.toString(j);
               if( depth != 1 ) tmp += "/";
               newList.add(tmp);
            }
         }
         strList.addAll(newList);
         oldList = newList;
         depth--;
      }

      log("Nodes generated: " +strList.size());
      return strList;
   }

   public static Test suite() throws Exception {
      return new TestSuite(ReplicatedSyncPerfTestCase.class);
   }

   private void log(String str) {
//     System.out.println(this.getClass().getName() +": " +str);
     System.out.println(str);
   }

}
