/*
 *
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.test.cache.test.replicated;

import EDU.oswego.cs.dl.util.concurrent.FIFOSemaphore;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import org.jboss.cache.TreeCache;
import org.jboss.cache.PropertyConfigurator;
import org.jboss.cache.lock.IsolationLevel;
import org.jboss.cache.transaction.DummyTransactionManager;
import org.jboss.logging.Logger;

import javax.naming.Context;
import javax.transaction.NotSupportedException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;

/**
 * Replicated unit test for sync transactional TreeCache
 * Note: we use DummyTransactionManager for Tx purpose instead of relying on
 * jta.
 *
 * @version $Revision: 1.5.2.2 $
 */
public class SyncTxUnitTestCase extends TestCase
{
   TreeCache cache1, cache2;
   int caching_mode = TreeCache.REPL_SYNC;
   final String group_name = "TreeCacheTestGroup";
   String props =
         "UDP(ip_mcast=true;ip_ttl=64;loopback=false;mcast_addr=228.1.2.3;" +
         "mcast_port=45566;mcast_recv_buf_size=80000;mcast_send_buf_size=150000;" +
         "ucast_recv_buf_size=80000;ucast_send_buf_size=150000):" +
         "PING(down_thread=true;num_initial_members=2;timeout=2000;up_thread=true):" +
         "MERGE2(max_interval=20000;min_interval=10000):" +
         "FD(down_thread=true;shun=true;up_thread=true):" +
         "VERIFY_SUSPECT(down_thread=true;timeout=1500;up_thread=true):" +
         "pbcast.NAKACK(down_thread=true;gc_lag=50;retransmit_timeout=600,1200,2400,4800;" +
         "up_thread=true):" +
         "pbcast.STABLE(desired_avg_gossip=20000;down_thread=true;up_thread=true):" +
         "UNICAST(down_thread=true;min_threshold=10;timeout=600,1200,2400;window_size=100):" +
         "FRAG(down_thread=true;frag_size=8192;up_thread=true):" +
         "pbcast.GMS(join_retry_timeout=2000;join_timeout=5000;print_local_addr=true;shun=true):" +
         "pbcast.STATE_TRANSFER(down_thread=true;up_thread=true)";

   final static Logger log_ = Logger.getLogger(SyncTxUnitTestCase.class);
   String old_factory = null;
   final String FACTORY = "org.jboss.cache.transaction.DummyContextFactory";
   static FIFOSemaphore lock_ = new FIFOSemaphore(1);
   DummyTransactionManager tx_mgr;


   public SyncTxUnitTestCase(String name)
   {
      super(name);
   }

   public void setUp() throws Exception
   {
      super.setUp();

      old_factory = System.getProperty(Context.INITIAL_CONTEXT_FACTORY);
      System.setProperty(Context.INITIAL_CONTEXT_FACTORY, FACTORY);

      tx_mgr = DummyTransactionManager.getInstance();

      initCaches(TreeCache.REPL_SYNC);
   }

   public void tearDown() throws Exception
   {
      super.tearDown();

      DummyTransactionManager.destroy();
      destroyCaches();

      if (old_factory != null) {
         System.setProperty(Context.INITIAL_CONTEXT_FACTORY, old_factory);
         old_factory = null;
      }
   }

   Transaction beginTransaction() throws SystemException, NotSupportedException
   {
      tx_mgr.begin();
      return tx_mgr.getTransaction();
   }

   void initCaches(int caching_mode) throws Exception
   {
      this.caching_mode = caching_mode;
      cache1 = new TreeCache(group_name, props, 3000);
      cache2 = new TreeCache(group_name, props, 3000);
      PropertyConfigurator config = new PropertyConfigurator();
      config.configure(cache1, "META-INF/replSync-service.xml"); // read in generic replAsync xml
      config.configure(cache2, "META-INF/replSync-service.xml"); // read in generic replAsync xml
      cache1.setCacheMode(caching_mode);
      cache2.setCacheMode(caching_mode);
      cache1.setIsolationLevel(IsolationLevel.REPEATABLE_READ);
      cache2.setIsolationLevel(IsolationLevel.REPEATABLE_READ);
      cache1.setTransactionManagerLookupClass("org.jboss.cache.JBossTransactionManagerLookup");
      cache2.setTransactionManagerLookupClass("org.jboss.cache.JBossTransactionManagerLookup");
      cache1.setLockAcquisitionTimeout(5000);
      cache2.setLockAcquisitionTimeout(5000);
      cache1.start();
      cache2.start();
   }

   void destroyCaches() throws Exception
   {
      cache1.stop();
      cache2.stop();
      cache1 = null;
      cache2 = null;
   }


   public void testConcurrentPuts() throws Exception
   {
      final TreeCache cache1 = this.cache1;

      Thread t1 = new Thread()
      {
         Transaction tx;

         public void run()
         {
            try {
               tx = beginTransaction();
               cache1.put("/bela/ban", "name", "Bela Ban");
               _pause(4000); // Thread2 will be blocked until we commit
               tx.commit();
            } catch (Exception ex) {
               fail(ex.toString());
            }
         }
      };

      Thread t2 = new Thread()
      {
         Transaction tx;

         public void run()
         {
            try {
               _pause(1000); // give Thread1 time to acquire the lock
               tx = beginTransaction();
               cache1.put("/bela/ban", "name", "Michelle Ban");
               tx.commit();
            } catch (Exception ex) {
               fail(ex.toString());
            }
         }
      };

      // Let the game start
      t1.start();
      t2.start();

      // Wait for thread to die but put an insurance of 5 seconds on it.
      t1.join();
      t2.join();

      assertEquals("Michelle Ban", cache1.get("/bela/ban", "name"));
   }

   public void testPut() throws Exception
   {
      final TreeCache cache1 = this.cache1;
      final TreeCache cache2 = this.cache2;
      Thread t1 = new Thread()
      {
         public void run()
         {
            try {
               lock_.acquire();
               cache1.put("/a/b/c", "age", new Integer(38));
               lock_.release();
               _pause(20);
               Thread.yield();

               lock_.acquire();
               cache1.put("/a/b/c", "age", new Integer(39));
               lock_.release();
               assertEquals(new Integer(39), cache1.get("/a/b/c", "age"));
            } catch (Exception ex) {
               fail(ex.toString());
            }
         }
      };

      Thread t2 = new Thread()
      {
         public void run()
         {
            try {
               _pause(10);
               Thread.yield();
               lock_.acquire();
               // Should replicate the value right away.
               Integer val = (Integer) cache2.get("/a/b/c", "age");
               assertEquals(new Integer(38), val);
               lock_.release();
               _pause(20);
               Thread.yield();

               lock_.acquire();
               val = (Integer) cache2.get("/a/b/c", "age");
               lock_.release();
               assertEquals(new Integer(39), val);
            } catch (Exception ex) {
               fail(ex.toString());
            }
         }
      };

      // Let the game start
      t1.start();
      t2.start();

      // Wait for thread to die but put an insurance of 5 seconds on it.
      t1.join();
      t2.join();
   }

   /**
    * Test replicated cache with transaction. Idea is to have two threads running
    * a local cache each that is replicating. Depending on whether cache1 commit/rollback or not,
    * the cache2.get will get different values.
    * Note that we have used sleep to interpose thread execution sequence.
    * Although it's not fool proof, it is rather simple and intuitive.
    *
    * @throws Exception
    */
   public void testPutTx() throws Exception
   {
      final TreeCache cache1 = this.cache1;
      final TreeCache cache2 = this.cache2;
      Thread t1 = new Thread()
      {
         public void run()
         {
            Transaction tx = null;

            try {
               lock_.acquire();
               tx = beginTransaction();
               cache1.put("/a/b/c", "age", new Integer(38));
               cache1.put("/a/b/c", "age", new Integer(39));
               lock_.release();
               _pause(100);
               lock_.acquire();
               tx.commit();
               lock_.release();
            } catch (Exception ex) {
               fail(ex.toString());
            }
         }
      };

      Thread t2 = new Thread()
      {
         public void run()
         {
            Transaction tx = null;

            try {
               sleep(20);
               Thread.yield();
               lock_.acquire();
               tx = beginTransaction();
               Object val = cache2.get("/a/b/c", "age"); // must be null as not yet committed
               assertTrue(val == null);
               tx.commit();
               lock_.release();

               _pause(100);
               lock_.acquire();
               tx = beginTransaction();
               assertEquals(new Integer(39), cache2.get("/a/b/c", "age")); // must not be null
               tx.commit();
               lock_.release();
            } catch (Exception ex) {
               fail(ex.toString());
            }
         }
      };

      // Let the game start
      t1.start();
      t2.start();

      // Wait for thread to die but put an insurance of 5 seconds on it.
      t1.join();
      t2.join();
   }

   /**
    * Have both cache1 and cache2 do add and commit. cache1 commit should time out
    * since it can't obtain the lock when trying to replicate cache2. On the other hand,
    * cache2 commit will succeed since now that cache1 is rollbacked and lock is
    * release.
    */
   public void testPutTx1() throws Exception
   {
      final TreeCache cache1 = this.cache1;
      final TreeCache cache2 = this.cache2;
      Thread t1 = new Thread()
      {
         public void run()
         {
            Transaction tx = null;

            try {
               lock_.acquire();
               tx = beginTransaction();
               cache1.put("/a/b/c", "age", new Integer(38));
               cache1.put("/a/b/c", "age", new Integer(39));
               lock_.release();

               _pause(100);
               lock_.acquire();
               try {
                  tx.commit();
               } catch (Exception tex) {
                  tex.printStackTrace();
                  return;
               } finally {
                  lock_.release();
               }
            } catch (Exception ex) {
               fail(ex.toString());
            }
         }
      };

      Thread t2 = new Thread()
      {
         public void run()
         {
            Transaction tx = null;

            try {
               sleep(20);
               Thread.yield();
               lock_.acquire();
               tx = beginTransaction();
               assertNull(cache2.get("/a/b/c", "age")); // must be null as not yet committed
               cache2.put("/a/b/c", "age", new Integer(40));
               lock_.release();

               _pause(100);
               lock_.acquire();
               assertEquals(new Integer(40), cache2.get("/a/b/c", "age")); // must not be null
               tx.commit();
               lock_.release();

               _pause(1000);
               tx = beginTransaction();
               assertEquals("After cache2 commit", new Integer(40), cache2.get("/a/b/c", "age"));
               tx.commit();
            } catch (Throwable ex) {
               ex.printStackTrace();
               fail(ex.toString());
            }
         }
      };

      // Let the game start
      t1.start();
      t2.start();

      t1.join();
      t2.join();
   }

   public void testPutTxWithRollback() throws Exception
   {
      final TreeCache cache1 = this.cache1;
      final TreeCache cache2 = this.cache2;
      Thread t1 = new Thread()
      {
         public void run()
         {
            Transaction tx = null;

            try {
               lock_.acquire();
               tx = beginTransaction();
               cache1.put("/a/b/c", "age", new Integer(38));
               cache1.put("/a/b/c", "age", new Integer(39));
               lock_.release();

               _pause(100);
               lock_.acquire();
               tx.rollback();
               lock_.release();
            } catch (Exception ex) {
               fail(ex.toString());
            }
         }
      };

      Thread t2 = new Thread()
      {
         public void run()
         {
            Transaction tx = null;

            try {
               sleep(20);
               Thread.yield();
               lock_.acquire();
               tx = beginTransaction();
               assertNull(cache2.get("/a/b/c", "age")); // must be null as not yet committed
               lock_.release();

               _pause(100);
               lock_.acquire();
               assertNull(cache2.get("/a/b/c", "age")); // must be null as rolledback
               tx.commit();
               lock_.release();
            } catch (Exception ex) {
               fail(ex.toString());
            }
         }
      };

      // Let the game start
      t1.start();
      t2.start();

      // Wait for thread to die but put an insurance of 5 seconds on it.
      t1.join();
      t2.join();
   }

   static void _pause(long millis)
   {
      try {
         Thread.sleep(millis);
      } catch (Exception ex) {
      }
   }

   public static Test suite() throws Exception
   {
//        return getDeploySetup(SyncTxUnitTestCase.class, "cachetest.jar");
      return new TestSuite(SyncTxUnitTestCase.class);
   }


}
