/*
 * JBoss, the OpenSource WebOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.web.tomcat.tc5;

import org.jboss.deployment.DeploymentException;
import org.jboss.deployment.DeploymentInfo;
import org.jboss.metadata.WebMetaData;
import org.jboss.util.file.JarUtils;
import org.jboss.web.AbstractWebContainer;
import org.jboss.web.AbstractWebDeployer;
import org.jboss.web.WebApplication;
import org.jboss.web.tomcat.security.JBossSecurityMgrRealm;
import org.jboss.web.tomcat.security.SecurityAssociationValve;
import org.jboss.web.tomcat.tc5.session.ClusteredSessionValve;
import org.jboss.web.tomcat.tc5.session.ClusteringNotSupportedException;
import org.jboss.web.tomcat.tc5.session.InstantSnapshotManager;
import org.jboss.web.tomcat.tc5.session.IntervalSnapshotManager;
import org.jboss.web.tomcat.tc5.session.JBossManager;
import org.jboss.web.tomcat.tc5.session.SnapshotManager;

import javax.management.Attribute;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URL;
import java.util.Iterator;
import java.util.zip.ZipFile;

/** The tomcat web application deployer
 * 
 * @author Scott.Stark@jboss.org
 * @author Costin Manolache
 * @version $Revision: 1.1.2.13 $
 */
public class TomcatDeployer extends AbstractWebDeployer
{
   private DeployerConfig config;
   private String[] javaVMs = 
   { " jboss.management.local:J2EEServer=Local,j2eeType=JVM,name=localhost" };
   private String serverName = "jboss";

   public void init(Object containerConfig) throws Exception
   {
      this.config = (DeployerConfig) containerConfig;
      super.setJava2ClassLoadingCompliance(config.java2ClassLoadingCompliance);
      super.setUnpackWars(config.unpackWars);
      super.setLenientEjbLink(config.lenientEjbLink);
   }

   /** 
    * Perform the tomcat specific deployment steps.
    */
   protected void performDeploy(WebApplication appInfo, String warUrl,
      AbstractWebContainer.WebDescriptorParser webAppParser)
      throws Exception
   {
      WebMetaData metaData = appInfo.getMetaData();
      String hostName = null;
      Iterator hostNames = metaData.getVirtualHosts();
      if (hostNames.hasNext())
      {
          hostName = hostNames.next().toString();
      }
      performDeployInternal(hostName, appInfo, warUrl, webAppParser);
      while (hostNames.hasNext())
      {
          String additionalHostName = hostNames.next().toString();
          performDeployInternal(additionalHostName, appInfo, warUrl, webAppParser);
      }
   }

   protected void performDeployInternal(String hostName, 
      WebApplication appInfo, String warUrl, 
      AbstractWebContainer.WebDescriptorParser webAppParser)
      throws Exception
   {

      WebMetaData metaData = appInfo.getMetaData();
      String ctxPath = metaData.getContextRoot();
      if (ctxPath.equals("/") || ctxPath.equals("/ROOT"))
      {
         log.debug("deploy root context=" + ctxPath);
         ctxPath = "";
         metaData.setContextRoot(ctxPath);
      }
      log.info("deploy, ctxPath=" + ctxPath + ", warUrl=" + warUrl);

      URL url = new URL(warUrl);

      // Servlet engines needs a war in a dir so extract the nested war
      if (url.getProtocol().equals("njar"))
      {
         url = org.jboss.net.protocol.njar.Handler.njarToFile(url);
         log.debug("Extracted war from njar, warUrl=" + url);
         File warFile = new File(url.getFile());
         String warFileName = warFile.getName();
         warFileName = warFileName.substring(0, warFileName.length() - 3);
         warFileName += "war";
         File warDir = new File(warFile.getParent(), warFileName);
         FileInputStream warStream = new FileInputStream(warFile);
         JarUtils.unjar(warStream, warDir);
         warStream.close();
         log.debug("Unpacked war into dir: " + warDir);
         url = warDir.toURL();
      }

      ClassLoader loader = Thread.currentThread().getContextClassLoader();

      // We need to establish the JNDI ENC prior to the start 
      // of the web container so that init on startup servlets are able 
      // to interact with their ENC. We hook into the context lifecycle 
      // events to be notified of the start of the
      // context as this occurs before the servlets are started.
      if (appInfo.getAppData() == null)
          webAppParser.parseWebAppDescriptors(loader, appInfo.getMetaData());

      appInfo.setName(url.getPath());
      appInfo.setClassLoader(loader);
      appInfo.setURL(url);

      if (ctxPath.equals(""))
         ctxPath = "/";
      if (ctxPath.equals("/ROOT"))
         ctxPath = "/";

      String objectNameS = config.getCatalinaDomain() 
          + ":j2eeType=WebModule,name=//" +
         ((hostName == null) ? "localhost" : hostName) 
          + ctxPath + ",J2EEApplication=none,J2EEServer=none";

      ObjectName objectName = new ObjectName(objectNameS);

      if (server.isRegistered(objectName))
      {
         log.debug("Already exists, destroying " + objectName);
         server.invoke(objectName, "destroy", new Object[]{},
            new String[]{});
      }

      server.createMBean("org.apache.commons.modeler.BaseModelMBean",
         objectName, new Object[]{config.getContextClassName()},
         new String[]{"java.lang.String"});

      server.setAttribute(objectName, new Attribute("docBase", url.getFile()));

      server.setAttribute(objectName, new Attribute
         ("defaultWebXml", "web.xml"));

      server.setAttribute(objectName, new Attribute("javaVMs", javaVMs));

      server.setAttribute(objectName, new Attribute("server", serverName));

      JBossSecurityMgrRealm jbossRealm = new JBossSecurityMgrRealm();
      jbossRealm.setSubjectAttributeName(config.getSubjectAttributeName());
      server.setAttribute(objectName, new Attribute("realm", jbossRealm));

      if (config.isUseJBossWebLoader())
      {
         WebCtxLoader webLoader = new WebCtxLoader(loader);
         webLoader.setWarURL(url);
         server.setAttribute(objectName, new Attribute
            ("loader", webLoader));
      }
      else
      {
         server.setAttribute(objectName, new Attribute
            ("parentClassLoader", loader));
      }

      server.setAttribute(objectName, new Attribute
         ("delegate", new Boolean(getJava2ClassLoadingCompliance())));

      String[] jspCP = getCompileClasspath(loader);
      StringBuffer classpath = new StringBuffer();
      for (int u = 0; u < jspCP.length; u++)
      {
         String repository = jspCP[u];
         if (repository == null)
            continue;
         if (repository.startsWith("file://"))
            repository = repository.substring(7);
         else if (repository.startsWith("file:"))
            repository = repository.substring(5);
         else
            continue;
         if (repository == null)
            continue;
         // ok it is a file.  Make sure that is is a directory or jar file
         File fp = new File(repository);
         if (!fp.isDirectory())
         {
            // if it is not a directory, try to open it as a zipfile.
            try
            {
               ZipFile zip = new ZipFile(fp);
            }
            catch (IOException e)
            {
               continue;
            }

         }
         if (u > 0)
            classpath.append(File.pathSeparator);
         classpath.append(repository);
      }

      server.setAttribute(objectName, new Attribute
         ("compilerClasspath", classpath.toString()));

      if (metaData.getDistributable())
      {
         // Try to initate clustering, fallback to standard if no clustering is available
         try
         {
            JBossManager manager = null;
            String managerClassName = config.getManagerClass();
            Class managerClass=Thread.currentThread().getContextClassLoader().loadClass(managerClassName);
            manager = (JBossManager)managerClass.newInstance();
            manager.init(ctxPath, this.log, metaData, config.isUseLocalCache());
            server.setAttribute(objectName, new Attribute("manager", manager));

            // todo (bela): get rid of the if-then-else stmt, use polymorphism (e.g. init()) to have each
            if(manager instanceof JBossManager)
            {
               // choose the snapshot manager
               SnapshotManager snap = null;
               String snapshotMode = config.getSnapshotMode();
               int snapshotInterval = config.getSnapshotInterval();
               if (snapshotMode.equals("instant"))
               {
                  snap = new InstantSnapshotManager(manager, ctxPath);
               }
               else if (snapshotMode.equals("interval"))
               {
                  snap = new IntervalSnapshotManager(manager, ctxPath, snapshotInterval);
               }
               else
               {
                  log.error("Snapshot mode must be 'instant' or 'interval' - using 'instant'");
                  snap = new InstantSnapshotManager(manager, ctxPath);
               }
               // Adding session snapshot valve
               Object valve = new ClusteredSessionValve(snap);
               server.invoke(objectName, "addValve",
                       new Object[]{valve},
                       new String[]{"org.apache.catalina.Valve"});
            }
            else
            {
               throw new ClusteringNotSupportedException("managerClass " + managerClassName + " not known");
            }

            log.debug("Enabled clustering support for ctxPath=" + ctxPath);
         }
         catch (ClusteringNotSupportedException e)
         {
            log.error("Failed to setup clustering, clustering disabled");
         }
      }

      // Set the session cookies flag according to metadata
      switch(metaData.getSessionCookies())
      {
         case WebMetaData.SESSION_COOKIES_ENABLED:
            server.setAttribute(objectName, new Attribute
               ("cookies", new Boolean(true)));
            log.debug("Enabling session cookies");
            break;
         case WebMetaData.SESSION_COOKIES_DISABLED:
            server.setAttribute(objectName, new Attribute
               ("cookies", new Boolean(false)));
            log.debug("Disabling session cookies");
            break;
         default:
            log.debug("Using session cookies default setting");
      }

      // Init the container; this will also start it
      server.invoke(objectName, "init", new Object[]{}, new String[]{});

      /* Add JBoss realm as valve after the externalized container valves so
       since this and the SecurityAssociationValve should be after any
       authentication/authorization side effects
      */
      server.invoke(objectName, "addValve",
         new Object[]{jbossRealm},
         new String[]{"org.apache.catalina.Valve"});

      // Add security association valve
      Object valve = new SecurityAssociationValve();
      server.invoke(objectName, "addValve",
         new Object[]{valve},
         new String[]{"org.apache.catalina.Valve"});

      // Retrieve the state, and throw an exception in case of a failure
      Integer state = (Integer) server.getAttribute(objectName, "state");
      if (state.intValue() != 1)
      {
         throw new DeploymentException("URL " + warUrl + " deployment failed");
      }

      // Ensure that any Authenticator in the context's pipeline requires 
      // that each request be reauthenticated to the security mgr
      ObjectName queryObjectName = new ObjectName
         (config.getCatalinaDomain() + ":host="
         + ((hostName == null) ? "localhost" : hostName)
         + ",path=" + ctxPath + ",type=Valve,*");

      Iterator iterator =
         server.queryMBeans(queryObjectName, null).iterator();

      while (iterator.hasNext())
      {
         ObjectName valveObjectName =
            ((ObjectInstance) iterator.next()).getObjectName();
         String name = valveObjectName.getKeyProperty("name");
         if ((name != null) && (name.indexOf("Authenticator") >= 0))
         {
            log.debug("Set auth base caching off on " + valveObjectName);
            server.setAttribute(valveObjectName, new Attribute
               ("cache", Boolean.FALSE));
         }
      }

      appInfo.setAppData(objectName);

      // Create mbeans for the servlets
      DeploymentInfo di = webAppParser.getDeploymentInfo();
      di.deployedObject = objectName;
      ObjectName servletQuery = new ObjectName
         (config.getCatalinaDomain() + ":j2eeType=Servlet,WebModule="
         + objectName.getKeyProperty("name") + ",*");
      iterator = server.queryMBeans(servletQuery, null).iterator();
      while (iterator.hasNext())
      {
         di.mbeans.add(((ObjectInstance) iterator.next()).getObjectName());
      }

      log.debug("Initialized: " + appInfo + " " + objectName);

   }


   /** Called as part of the undeploy() method template to ask the
    subclass for perform the web container specific undeployment steps.
    */
   protected void performUndeploy(String warUrl, WebApplication appInfo)
      throws Exception
   {
      if (appInfo == null)
      {
         log.debug("performUndeploy, no WebApplication found for URL "
            + warUrl);
         return;
      }

      log.info("undeploy, ctxPath=" + appInfo.getMetaData().getContextRoot()
         + ", warUrl=" + warUrl);

      WebMetaData metaData = appInfo.getMetaData();
      String hostName = null;
      Iterator hostNames = metaData.getVirtualHosts();
      if (hostNames.hasNext()) {
          hostName = hostNames.next().toString();
      }
      performUndeployInternal(hostName, warUrl, appInfo);
      while (hostNames.hasNext()) {
          String additionalHostName = hostNames.next().toString();
          performUndeployInternal(additionalHostName, warUrl, appInfo);
      }

   }

   protected void performUndeployInternal(String hostName, String warUrl, 
      WebApplication appInfo)
      throws Exception
   {

      WebMetaData metaData = appInfo.getMetaData();
      String ctxPath = metaData.getContextRoot();

      // If the server is gone, all apps were stopped already
      if (server == null)
         return;

      ObjectName objectName = new ObjectName(config.getCatalinaDomain() 
          + ":j2eeType=WebModule,name=//" +
         ((hostName == null) ? "localhost" : hostName) 
          + ctxPath + ",J2EEApplication=none,J2EEServer=none");

      if (server.isRegistered(objectName))
      {
         // Contexts should be stopped by the host already
         server.invoke(objectName, "destroy", new Object[]{},
            new String[]{});
      }

   }


}
